import re
from typing import List, Optional

from rbx.providers.provider import ProviderInterface
from rbx.schema import Problem


def _add_underscores(matches: List[str]) -> List[str]:
    if len(matches) <= 2:
        return matches
    return [f'{x}_' for x in matches[:-2]] + matches[-2:]


_PATTERNS = [
    r'https?://(?:.*\.)?codeforces.(?:com|ml|es)/(?:contest|gym)/(\d+)/problem/([^/]+)',
    r'https?://(?:.*\.)?codeforces.(?:com|ml|es)/problemset/problem/(\d+)/([^/]+)',
    (
        r'https?://(?:.*\.)?codeforces.(?:com|ml|es)/group/([^/]+)/contest/(\d+)/problem/([^/]+)',
        _add_underscores,
    ),
    r'https?://(?:.*\.)?codeforces.(?:com|ml|es)/problemset/(gym)Problem/([^/]+)',
    r'https?://(?:.*\.)?codeforces.(?:com|ml|es)/problemsets/(acm)sguru/problem/(?:\d+)/([^/]+)',
    # TODO: add EDU
]


def _compiled_pattern(pattern):
    if isinstance(pattern, tuple):
        return (re.compile(pattern[0]), pattern[1])
    return re.compile(pattern)


_COMPILED_PATTERNS = [_compiled_pattern(pattern) for pattern in _PATTERNS]


def get_code_tuple(url: str) -> Optional[List[str]]:
    for pattern_obj in _COMPILED_PATTERNS:
        if isinstance(pattern_obj, tuple):
            pattern, extract = pattern_obj
        else:
            pattern = pattern_obj
            extract = lambda x: x  # noqa: E731

        if match := pattern.match(url):
            return extract(list(match.groups()))
    return None


class CodeforcesProvider(ProviderInterface):
    def should_handle(self, url: str) -> bool:
        return 'codeforces.com/' in url

    def should_simplify_contest_problems(self) -> bool:
        return True

    def get_problem_code_within_contest(self, problem: Problem) -> str:
        return self.get_aliases(problem)[-1]

    def get_code(self, problem: Problem) -> str:
        code_tuple = get_code_tuple(problem.url)
        if not code_tuple:
            return super().get_code(problem)
        return ''.join(code_tuple)

    def get_aliases(self, problem: Problem) -> List[str]:
        code_tuple = get_code_tuple(problem.url)
        if not code_tuple:
            return super().get_aliases(problem)

        aliases = []
        for i in range(len(code_tuple)):
            aliases.append(''.join(code_tuple[i:]))
        return aliases



import pathlib
from typing import Annotated, List, Optional

import syncer
import typer

from rbx import annotations, config, utils
from rbx.box import package
from rbx.box.generators import (
    GenerationTestcaseEntry,
    generate_outputs_for_testcases,
    generate_standalone,
)
from rbx.box.testcase_extractors import (
    extract_generation_testcases,
    extract_generation_testcases_from_patterns,
)
from rbx.box.testcase_utils import TestcaseEntry, TestcasePattern
from rbx.console import console

app = typer.Typer(no_args_is_help=True, cls=annotations.AliasGroup)


async def _find_testcase(entry: TestcaseEntry) -> GenerationTestcaseEntry:
    extracted = await extract_generation_testcases([entry])
    if not extracted:
        console.print(f'[error]Testcase [item]{entry}[/item] not found.[/error]')
        raise typer.Exit(1)
    return extracted[0]


def _should_generate_output(entry: GenerationTestcaseEntry) -> bool:
    return (
        entry.metadata.copied_from is None
        or entry.metadata.copied_from.outputPath is None
    ) and package.get_main_solution() is not None


async def _generate_input_for_editing(
    entry: GenerationTestcaseEntry,
    output: bool = True,
    progress: Optional[utils.StatusProgress] = None,
) -> pathlib.Path:
    if (
        output and _should_generate_output(entry)
    ) or entry.metadata.copied_from is None:
        await generate_standalone(
            entry.metadata,
            validate=False,
            group_entry=entry.group_entry,
            progress=progress,
        )
    if entry.metadata.copied_from is not None:
        return entry.metadata.copied_from.inputPath
    return entry.metadata.copied_to.inputPath


async def _generate_output_for_editing(
    entry: GenerationTestcaseEntry,
    progress: Optional[utils.StatusProgress] = None,
) -> Optional[pathlib.Path]:
    if (
        entry.metadata.copied_from is not None
        and entry.metadata.copied_from.outputPath is not None
    ):
        return entry.metadata.copied_from.outputPath
    if not _should_generate_output(entry):
        return None
    await generate_outputs_for_testcases([entry.group_entry], progress=progress)
    return entry.metadata.copied_to.outputPath


async def _generate_for_editing(
    entry: GenerationTestcaseEntry,
    input: bool,
    output: bool,
    progress: Optional[utils.StatusProgress] = None,
) -> List[pathlib.Path]:
    res = []
    input_path = await _generate_input_for_editing(
        entry, output=output, progress=progress
    )
    if input:
        res.append(input_path)
    if output:
        output_path = await _generate_output_for_editing(entry, progress=progress)
        if output_path is not None:
            res.append(output_path)
    return res


@app.command('view, v', help='View a testcase in your default editor.')
@syncer.sync
async def view(
    tc: Annotated[
        str,
        typer.Argument(help='Testcase to view. Format: [group]/[index].'),
    ],
    input_only: bool = typer.Option(
        False,
        '--input',
        '-i',
        help='Whether to open only the input file in the editor.',
    ),
    output_only: bool = typer.Option(
        False,
        '--output',
        '-o',
        help='Whether to open only the output file in the editor.',
    ),
):
    if input_only and output_only:
        console.print(
            '[error]Flags --input and --output cannot be used together.[/error]'
        )
        raise typer.Exit(1)

    entry = TestcaseEntry.parse(tc)
    testcase = await _find_testcase(entry)

    with utils.StatusProgress('Preparing testcase...') as s:
        items = await _generate_for_editing(
            testcase, input=not output_only, output=not input_only, progress=s
        )
    config.edit_multiple(items, readonly=True)


@app.command('info, i', help='Show information about testcases.')
@syncer.sync
async def info(
    pattern: Annotated[
        Optional[str],
        typer.Argument(
            help='Testcases to detail, as a pattern. Might be a group, or a specific test in the format [group]/[index].'
        ),
    ] = None,
):
    tc_pattern = TestcasePattern.parse(pattern or '*')
    testcases = await extract_generation_testcases_from_patterns([tc_pattern])
    if not testcases:
        console.print(
            f'[error]No testcases found matching pattern [item]{pattern}[/item].[/error]'
        )
        raise typer.Exit(1)

    for testcase in testcases:
        console.print(f'[status]Identifier:[/status] {testcase.group_entry}')
        if testcase.metadata.generator_call is not None:
            console.print(
                f'[status]Generator call:[/status] {testcase.metadata.generator_call}'
            )
        if testcase.metadata.copied_from is not None:
            console.print(
                f'[status]Input file:[/status] {testcase.metadata.copied_from.inputPath}'
            )
            if testcase.metadata.copied_from.outputPath is not None:
                console.print(
                    f'[status]Output file:[/status] {testcase.metadata.copied_from.outputPath}'
                )

        if testcase.metadata.generator_script is not None:
            console.print(
                f'[status]Generator script:[/status] {testcase.metadata.generator_script.path}, line {testcase.metadata.generator_script.line}'
            )
        console.print()



# type: ignore
from enum import Enum, auto
from typing import Any, Dict, List, Optional, Set, Tuple, Union


class alias(auto):
    def __init__(self, *aliases):
        if len(aliases) == 0:
            raise ValueError('Cannot have empty alias() call.')
        for a in aliases:
            if not isinstance(a, str):
                raise ValueError(
                    f'All aliases for must be strings; found alias of type {type(a)} having value: {a}'
                )
        self.names = aliases
        self.enum_name = None

    def __repr__(self) -> str:
        return str(self)

    def __str__(self):
        if self.enum_name is not None:
            return self.enum_name
        return self.alias_repr

    @property
    def alias_repr(self) -> str:
        return str(f'alias:{list(self.names)}')

    def __setattr__(self, attr_name: str, attr_value: Any):
        if attr_name == 'value':
            ## because alias subclasses auto and does not set value, enum.py:143 will try to set value
            self.enum_name = attr_value
        else:
            super(alias, self).__setattr__(attr_name, attr_value)

    def __getattribute__(self, attr_name: str):
        """
        Refer these lines in Python 3.10.9 enum.py:

        class _EnumDict(dict):
            ...
            def __setitem__(self, key, value):
                ...
                elif not _is_descriptor(value):
                    ...
                    if isinstance(value, auto):
                        if value.value == _auto_null:
                            value.value = self._generate_next_value(
                                    key,
                                    1,
                                    len(self._member_names),
                                    self._last_values[:],
                                    )
                            self._auto_called = True
                        value = value.value
                    ...
                ...
            ...

        """
        if attr_name == 'value':
            if object.__getattribute__(self, 'enum_name') is None:
                ## Gets _auto_null as alias inherits auto class but does not set `value` class member; refer enum.py:142
                try:
                    return object.__getattribute__(self, 'value')
                except Exception:
                    from enum import _auto_null

                    return _auto_null
            return self
        return object.__getattribute__(self, attr_name)


_DEFAULT_REMOVAL_TABLE = str.maketrans(
    'ABCDEFGHIJKLMNOPQRSTUVWXYZ',
    'abcdefghijklmnopqrstuvwxyz',
    ' -_.:;,',  ## Will be removed
)


class AutoEnum(str, Enum):
    """
    Utility class which can be subclassed to create enums using auto() and alias().
    Also provides utility methods for common enum operations.
    """

    def __init__(self, value: Union[str, alias]):
        self.aliases: Tuple[str, ...] = tuple()
        if isinstance(value, alias):
            self.aliases: Tuple[str, ...] = value.names

    @classmethod
    def _missing_(cls, enum_value: Any):
        ## Ref: https://stackoverflow.com/a/60174274/4900327
        ## This is needed to allow Pydantic to perform case-insensitive conversion to AutoEnum.
        return cls.from_str(enum_value=enum_value, raise_error=True)

    def _generate_next_value_(name, start, count, last_values):
        return name

    @property
    def str(self) -> str:
        return self.__str__()

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash(self.__class__.__name__ + '.' + self.name)

    def __eq__(self, other):
        return self is other

    def __ne__(self, other):
        return self is not other

    def matches(self, enum_value: str) -> bool:
        return self is self.from_str(enum_value, raise_error=False)

    @classmethod
    def matches_any(cls, enum_value: str) -> bool:
        return cls.from_str(enum_value, raise_error=False) is not None

    @classmethod
    def does_not_match_any(cls, enum_value: str) -> bool:
        return not cls.matches_any(enum_value)

    @classmethod
    def display_names(cls, **kwargd) -> str:
        return str([enum_value.display_name(**kwargd) for enum_value in list(cls)])

    def display_name(self, *, sep: str = ' ') -> str:
        return sep.join(
            [
                (
                    word.lower()
                    if word.lower() in ('of', 'in', 'the')
                    else word.capitalize()
                )
                for word in str(self).split('_')
            ]
        )

    @classmethod
    def _initialize_lookup(cls):
        if (
            '_value2member_map_normalized_' not in cls.__dict__
        ):  ## Caching values for fast retrieval.
            cls._value2member_map_normalized_ = {}

            def _set_normalized(e, normalized_e_name):
                if normalized_e_name in cls._value2member_map_normalized_:
                    print('DEU MERDA')
                    raise ValueError(
                        f'Cannot register enum "{e.name}"; '
                        f'another enum with the same normalized name "{normalized_e_name}" already exists.'
                    )
                cls._value2member_map_normalized_[normalized_e_name] = e

            for e in list(cls):
                _set_normalized(e, cls._normalize(e.name))
                if len(e.aliases) > 0:
                    ## Add the alias-repr to the lookup:
                    _set_normalized(e, cls._normalize(alias(*e.aliases).alias_repr))
                    candidate_aliases = set(
                        cls._normalize(e_alias) for e_alias in e.aliases
                    )
                    if cls._normalize(e.name) in candidate_aliases:
                        candidate_aliases.remove(cls._normalize(e.name))
                    for e_alias in candidate_aliases:
                        _set_normalized(e, e_alias)

    @classmethod
    def from_str(cls, enum_value: str, raise_error: bool = True) -> Optional:
        """
        Performs a case-insensitive lookup of the enum value string among the members of the current AutoEnum subclass.
        :param enum_value: enum value string
        :param raise_error: whether to raise an error if the string is not found in the enum
        :return: an enum value which matches the string
        :raises: ValueError if raise_error is True and no enum value matches the string
        """
        if isinstance(enum_value, cls):
            return enum_value
        if enum_value is None and raise_error is False:
            return None
        if not isinstance(enum_value, str) and raise_error is True:
            raise ValueError(f'Input should be a string; found type {type(enum_value)}')
        cls._initialize_lookup()
        enum_obj: Optional[AutoEnum] = cls._value2member_map_normalized_.get(
            cls._normalize(enum_value)
        )
        if enum_obj is None and raise_error is True:
            raise ValueError(
                f'Could not find enum with value {repr(enum_value)}; available values are: {list(cls)}.'
            )
        return enum_obj

    @classmethod
    def _normalize(cls, x: str) -> str:
        ## Found to be faster than .translate() and re.sub() on Python 3.10.6
        return str(x).translate(_DEFAULT_REMOVAL_TABLE)

    @classmethod
    def convert_keys(cls, d: Dict) -> Dict:
        """
        Converts string dict keys to the matching members of the current AutoEnum subclass.
        Leaves non-string keys untouched.
        :param d: dict to transform
        :return: dict with matching string keys transformed to enum values
        """
        out_dict = {}
        for k, v in d.items():
            if isinstance(k, str) and cls.from_str(k, raise_error=False) is not None:
                out_dict[cls.from_str(k, raise_error=False)] = v
            else:
                out_dict[k] = v
        return out_dict

    @classmethod
    def convert_keys_to_str(cls, d: Dict) -> Dict:
        """
        Converts dict keys of the current AutoEnum subclass to the matching string key.
        Leaves other keys untouched.
        :param d: dict to transform
        :return: dict with matching keys of the current AutoEnum transformed to strings.
        """
        out_dict = {}
        for k, v in d.items():
            if isinstance(k, cls):
                out_dict[str(k)] = v
            else:
                out_dict[k] = v
        return out_dict

    @classmethod
    def convert_values(
        cls, d: Union[Dict, Set, List, Tuple], raise_error: bool = False
    ) -> Union[Dict, Set, List, Tuple]:
        """
        Converts string values to the matching members of the current AutoEnum subclass.
        Leaves non-string values untouched.
        :param d: dict, set, list or tuple to transform.
        :param raise_error: raise an error if unsupported type.
        :return: data structure with matching string values transformed to enum values.
        """
        if isinstance(d, dict):
            return cls.convert_dict_values(d)
        if isinstance(d, list):
            return cls.convert_list(d)
        if isinstance(d, tuple):
            return tuple(cls.convert_list(d))
        if isinstance(d, set):
            return cls.convert_set(d)
        if raise_error:
            raise ValueError(f'Unrecognized data structure of type {type(d)}')
        return d

    @classmethod
    def convert_dict_values(cls, d: Dict) -> Dict:
        """
        Converts string dict values to the matching members of the current AutoEnum subclass.
        Leaves non-string values untouched.
        :param d: dict to transform
        :return: dict with matching string values transformed to enum values
        """
        out_dict = {}
        for k, v in d.items():
            if isinstance(v, str) and cls.from_str(v, raise_error=False) is not None:
                out_dict[k] = cls.from_str(v, raise_error=False)
            else:
                out_dict[k] = v
        return out_dict

    @classmethod
    def convert_list(cls, ls: Union[List, Tuple]) -> List:
        """
        Converts string list itmes to the matching members of the current AutoEnum subclass.
        Leaves non-string items untouched.
        :param l: list to transform
        :return: list with matching string items transformed to enum values
        """
        out_list = []
        for item in ls:
            if isinstance(item, str) and cls.matches_any(item):
                out_list.append(cls.from_str(item))
            else:
                out_list.append(item)
        return out_list

    @classmethod
    def convert_set(cls, s: Set) -> Set:
        """
        Converts string list itmes to the matching members of the current AutoEnum subclass.
        Leaves non-string items untouched.
        :param s: set to transform
        :return: set with matching string items transformed to enum values
        """
        out_set = set()
        for item in s:
            if isinstance(item, str) and cls.matches_any(item):
                out_set.add(cls.from_str(item))
            else:
                out_set.add(item)
        return out_set

    @classmethod
    def convert_values_to_str(cls, d: Dict) -> Dict:
        """
        Converts dict values of the current AutoEnum subclass to the matching string value.
        Leaves other values untouched.
        :param d: dict to transform
        :return: dict with matching values of the current AutoEnum transformed to strings.
        """
        out_dict = {}
        for k, v in d.items():
            if isinstance(v, cls):
                out_dict[k] = str(v)
            else:
                out_dict[k] = v
        return out_dict


class TestEnum(AutoEnum):
    ACCEPTED = alias('accepted', 'ac')
    WRONG_ANSWER = alias('wrong answer', 'wa')


if __name__ == '__main__':
    print(TestEnum('ac'))



import dataclasses
import pathlib
import string
from typing import List, Tuple

from rich.text import Text


@dataclasses.dataclass
class TruncatedOutput:
    truncate: bool = False
    lines: List[Tuple[int, str]] = dataclasses.field(default_factory=list)


def split_and_truncate_in_lines(
    s: str, max_line_length: int = 64, max_lines: int = 30
) -> TruncatedOutput:
    lines: List[Tuple[int, str]] = []
    current_line = []
    current_line_idx = 1

    def end_line(wrap: bool = False):
        nonlocal current_line, current_line_idx
        lines.append((current_line_idx, ''.join(current_line)))
        current_line = []
        if not wrap:
            current_line_idx += 1

    printable = set(string.printable)
    truncate = False
    for c in s:
        if c == '\n':
            end_line()
            continue
        if c not in printable:
            # TODO: handle better
            continue
        if len(current_line) >= max_line_length:
            end_line(wrap=True)
        if current_line_idx > max_lines:
            truncate = True
            break
        current_line.append(c)

    if current_line:
        end_line()

    return TruncatedOutput(truncate=truncate, lines=lines)


def _largest_line_number_length(lines: List[Tuple[int, str]]) -> int:
    return max([len(str(line[0])) for line in lines] + [1])


def render(s: str):
    truncated = split_and_truncate_in_lines(s)
    number_len = _largest_line_number_length(truncated.lines)

    text = Text()

    last_number = 0
    for line in truncated.lines:
        number, content = line
        number_str = '' if last_number == number else str(number)
        text.append(f'{number_str:>{number_len}}', style='lnumber')
        text.append(' ' * 3)
        text.append(content)
        text.append('\n')

        last_number = number
    if truncated.truncate:
        text.append(f"{'':>{number_len}}", style='lnumber')
        text.append(' ' * 3)
        text.append('... (truncated)')
    return text


def render_from_file(file: pathlib.Path):
    return render(file.read_text())



import dataclasses
import pathlib
import subprocess
from typing import Optional

import chardet

MAX_PDFLATEX_RUNS = 3


def should_rerun(logs: str) -> bool:
    logs = logs.lower()
    for line in logs.splitlines():
        if 'rerun to get cross-references right' in line:
            return True
        if 'rerun' in line and 'warning' in line:
            return True
    return False


def decode_latex_output(output: bytes) -> str:
    # Latex output can be tricky with decoding
    encoding = chardet.detect(output)['encoding'] or 'utf-8'
    return output.decode(encoding)


@dataclasses.dataclass
class LatexResult:
    result: subprocess.CompletedProcess[bytes]
    pdf: Optional[bytes]


class Latex:
    def __init__(self, latex: str):
        self.latex = latex

    def build_pdf(self, temp_dir: pathlib.Path) -> LatexResult:
        temp_path = temp_dir / 'statement.tex'
        output_path = temp_path.with_suffix('.pdf')
        args = ['pdflatex', '-interaction', 'nonstopmode', str(temp_path)]
        temp_path.write_text(self.latex)

        completed = subprocess.run(args, timeout=15, capture_output=True, cwd=temp_dir)
        if completed.returncode != 0 or not output_path.exists():
            return LatexResult(result=completed, pdf=None)

        return LatexResult(result=completed, pdf=output_path.read_bytes())



import contextlib
import fcntl
import functools
import json
import os
import pathlib
import resource
from typing import Any, Optional, Type, TypeVar

import rich
import rich.prompt
import rich.status
import ruyaml
import typer
import yaml
from pydantic import BaseModel
from rich import text
from rich.highlighter import JSONHighlighter

from rbx.console import console

T = TypeVar('T', bound=BaseModel)
APP_NAME = 'rbx'


def create_and_write(path: pathlib.Path, *args, **kwargs):
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(*args, **kwargs)


def highlight_str(s: str) -> text.Text:
    txt = text.Text(s)
    JSONHighlighter().highlight(txt)
    return txt


def highlight_json_obj(obj: Any) -> text.Text:
    js = json.dumps(obj)
    return highlight_str(js)


def normalize_with_underscores(s: str) -> str:
    res = s.replace(' ', '_').replace('.', '_').strip('_')
    final = []

    last = ''
    for c in res:
        if c == '_' and last == c:
            continue
        last = c
        final.append(c)
    return ''.join(final)


def get_app_path() -> pathlib.Path:
    app_dir = typer.get_app_dir(APP_NAME)
    return pathlib.Path(app_dir)


def ensure_schema(model: Type[BaseModel]) -> pathlib.Path:
    path = get_app_path() / 'schemas' / f'{model.__name__}.json'
    path.parent.mkdir(parents=True, exist_ok=True)
    schema = json.dumps(model.model_json_schema(), indent=4)
    path.write_text(schema)
    return path.resolve()


def model_json(model: BaseModel) -> str:
    ensure_schema(model.__class__)
    return model.model_dump_json(indent=4, exclude_unset=True, exclude_none=True)


def model_to_yaml(model: BaseModel) -> str:
    path = ensure_schema(model.__class__)
    return f'# yaml-language-server: $schema={path}\n\n' + yaml.dump(
        model.model_dump(mode='json', exclude_unset=True, exclude_none=True),
        sort_keys=False,
        allow_unicode=True,
    )


def model_from_yaml(model: Type[T], s: str) -> T:
    ensure_schema(model)
    return model(**yaml.safe_load(s))


def validate_field(model: Type[T], field: str, value: Any):
    model.__pydantic_validator__.validate_assignment(
        model.model_construct(), field, value
    )


def save_ruyaml(path: pathlib.Path, yml: ruyaml.YAML, data: ruyaml.Any):
    path.parent.mkdir(parents=True, exist_ok=True)
    with path.open('w') as f:
        yml.dump(data, f)


@functools.cache
def get_empty_sentinel_path() -> pathlib.Path:
    path = get_app_path() / '.empty'
    path.write_text('')
    return path


@contextlib.contextmanager
def no_progress(status: Optional[rich.status.Status]):
    if status:
        status.stop()
    yield
    if status:
        status.start()


def confirm_on_status(status: Optional[rich.status.Status], *args, **kwargs) -> bool:
    with no_progress(status):
        res = rich.prompt.Confirm.ask(*args, **kwargs, console=console)
    return res


def get_open_fds():
    fds = []
    soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
    for fd in range(0, soft):
        try:
            fcntl.fcntl(fd, fcntl.F_GETFD)
        except IOError:
            continue
        fds.append(fd)
    return fds


@contextlib.contextmanager
def new_cd(x: pathlib.Path):
    d = os.getcwd()

    # This could raise an exception, but it's probably
    # best to let it propagate and let the caller
    # deal with it, since they requested x
    os.chdir(x)

    try:
        yield

    finally:
        # This could also raise an exception, but you *really*
        # aren't equipped to figure out what went wrong if the
        # old working directory can't be restored.
        os.chdir(d)


class StatusProgress(rich.status.Status):
    _message: str
    processed: int
    keep: bool

    def __init__(
        self, message: str, formatted_message: Optional[str] = None, keep: bool = False
    ):
        self._message = formatted_message or message
        self.keep = keep
        self.processed = 0
        super().__init__(message.format(processed=0), console=console)
        self.start()

    def __enter__(self):
        super().__enter__()
        return self

    def __exit__(self, *args, **kwargs):
        super().__exit__(*args, **kwargs)
        if self.keep:
            console.print(self._message.format(processed=self.processed))

    def update_with_progress(self, processed: int):
        self.processed = processed
        self.update(self._message.format(processed=processed))

    def step(self, delta: int = 1):
        self.processed += delta
        self.update_with_progress(self.processed)



import pathlib
import shutil
import tempfile
from typing import Annotated, Iterable, List, Optional, Sequence, Union

import rich
import rich.prompt
import typer
from iso639.language import functools

from rbx import console, utils
from rbx.box import cd
from rbx.box.environment import get_environment_path
from rbx.box.presets.fetch import PresetFetchInfo, get_preset_fetch_info
from rbx.box.presets.lock_schema import LockedAsset, PresetLock
from rbx.box.presets.schema import Preset, TrackedAsset
from rbx.config import get_default_app_path
from rbx.grading.judge.digester import digest_cooperatively, digest_file

app = typer.Typer(no_args_is_help=True)

LOCAL = 'local'


def find_preset_yaml(root: pathlib.Path = pathlib.Path()) -> Optional[pathlib.Path]:
    found = root / 'preset.rbx.yml'
    if found.exists():
        return found
    return None


def get_preset_yaml(root: pathlib.Path = pathlib.Path()) -> Preset:
    found = find_preset_yaml(root)
    if not found:
        console.console.print(
            f'[error][item]preset.rbx.yml[/item] not found in [item]{root.absolute()}[/item][/error]'
        )
        raise typer.Exit(1)
    return utils.model_from_yaml(Preset, found.read_text())


def find_preset_lock(root: pathlib.Path = pathlib.Path()) -> Optional[pathlib.Path]:
    root = root.resolve()
    problem_yaml_path = root / '.preset-lock.yml'
    if not problem_yaml_path.is_file():
        return None
    return problem_yaml_path


def get_preset_lock(root: pathlib.Path = pathlib.Path()) -> Optional[PresetLock]:
    found = find_preset_lock(root)
    if not found:
        return None
    return utils.model_from_yaml(PresetLock, found.read_text())


def _find_nested_preset(root: pathlib.Path) -> Optional[pathlib.Path]:
    root = root.resolve()
    problem_yaml_path = root / 'preset.rbx.yml'
    while root != pathlib.PosixPath('/') and not problem_yaml_path.is_file():
        root = root.parent
        problem_yaml_path = root / 'preset.rbx.yml'
    if not problem_yaml_path.is_file():
        return None
    return problem_yaml_path.parent


def _find_local_preset(root: pathlib.Path) -> Optional[pathlib.Path]:
    original_root = root
    root = root.resolve()
    problem_yaml_path = root / '.local.rbx' / 'preset.rbx.yml'
    while root != pathlib.PosixPath('/') and not problem_yaml_path.is_file():
        root = root.parent
        problem_yaml_path = root / '.local.rbx' / 'preset.rbx.yml'
    if not problem_yaml_path.is_file():
        return _find_nested_preset(original_root)
    return problem_yaml_path.parent


def get_preset_installation_path(
    name: str, root: pathlib.Path = pathlib.Path()
) -> pathlib.Path:
    if name == LOCAL:
        local_path = _find_local_preset(root)
        if local_path is None:
            console.console.print('[error]Local preset was not found.[/error]')
            raise typer.Exit(1)
        return local_path
    nested_preset_path = _find_nested_preset(root)
    if nested_preset_path is not None:
        nested_preset = utils.model_from_yaml(
            Preset, (nested_preset_path / 'preset.rbx.yml').read_text()
        )
        if nested_preset.name == name:
            return nested_preset_path
    return utils.get_app_path() / 'presets' / name


def _find_installed_presets() -> List[str]:
    folder = utils.get_app_path() / 'presets'

    res = []
    if get_preset_installation_path('local'):
        res.append('local')
    for yml in folder.glob('*/preset.rbx.yml'):
        res.append(yml.parent.name)
    return res


def _is_contest(root: pathlib.Path = pathlib.Path()) -> bool:
    return (root / 'contest.rbx.yml').is_file()


def _is_problem(root: pathlib.Path = pathlib.Path()) -> bool:
    return (root / 'problem.rbx.yml').is_file()


def _check_is_valid_package(root: pathlib.Path = pathlib.Path()):
    if not _is_contest(root) and not _is_problem(root):
        console.console.print('[error]Not a valid rbx package directory.[/error]')
        raise typer.Exit(1)


def _get_preset_package_path(
    name: str, is_contest: bool, root: pathlib.Path = pathlib.Path()
) -> pathlib.Path:
    preset_path = get_preset_installation_path(name, root)
    preset = get_installed_preset(name, root)

    if is_contest:
        assert (
            preset.contest is not None
        ), 'Preset does not have a contest package definition.'
        return preset_path / preset.contest

    assert (
        preset.problem is not None
    ), 'Preset does not have a problem package definition,'
    return preset_path / preset.problem


def _process_globbing(
    assets: Iterable[TrackedAsset], preset_dir: pathlib.Path
) -> List[TrackedAsset]:
    res = []
    for asset in assets:
        if '*' in str(asset.path):
            glb = str(asset.path)
            files = preset_dir.glob(glb)
            relative_files = [file.relative_to(preset_dir) for file in files]
            res.extend([TrackedAsset(path=path) for path in relative_files])
            continue
        res.append(asset)
    return res


def _get_preset_tracked_assets(name: str, is_contest: bool) -> List[TrackedAsset]:
    preset = get_installed_preset(name)
    preset_path = get_preset_installation_path(name)

    if is_contest:
        assert (
            preset.contest is not None
        ), 'Preset does not have a contest package definition.'
        return _process_globbing(preset.tracking.contest, preset_path)

    assert (
        preset.problem is not None
    ), 'Preset does not have a problem package definition,'
    return _process_globbing(preset.tracking.problem, preset_path)


def _build_package_locked_assets(
    tracked_assets: Sequence[Union[TrackedAsset, LockedAsset]],
    root: pathlib.Path = pathlib.Path(),
) -> List[LockedAsset]:
    res = []
    for tracked_asset in tracked_assets:
        asset_path = root / tracked_asset.path
        if not asset_path.is_file():
            continue
        with asset_path.open('rb') as f:
            res.append(
                LockedAsset(path=tracked_asset.path, hash=digest_cooperatively(f))
            )
    return res


def _find_non_modified_assets(
    reference: List[LockedAsset], current: List[LockedAsset]
) -> List[LockedAsset]:
    current_by_path = {asset.path: asset for asset in current}

    res = []
    for asset in reference:
        if (
            asset.path in current_by_path
            and current_by_path[asset.path].hash != asset.hash
        ):
            # This is a file that was modified.
            continue
        res.append(asset)
    return res


def _find_modified_assets(
    reference: List[LockedAsset],
    current: List[LockedAsset],
):
    reference_by_path = {asset.path: asset for asset in reference}

    res = []
    for asset in current:
        if (
            asset.path in reference_by_path
            and reference_by_path[asset.path].hash == asset.hash
        ):
            # This is a file that was not modified.
            continue
        res.append(asset)
    return res


def _copy_updated_assets(
    preset_name: str,
    preset_lock: PresetLock,
    is_contest: bool,
    root: pathlib.Path = pathlib.Path(),
):
    current_package_snapshot = _build_package_locked_assets(preset_lock.assets)
    non_modified_assets = _find_non_modified_assets(
        preset_lock.assets, current_package_snapshot
    )

    preset_package_path = _get_preset_package_path(preset_name, is_contest=is_contest)
    preset_tracked_assets = _get_preset_tracked_assets(
        preset_name, is_contest=is_contest
    )
    current_preset_snapshot = _build_package_locked_assets(
        preset_tracked_assets, preset_package_path
    )
    assets_to_copy = _find_modified_assets(non_modified_assets, current_preset_snapshot)

    for asset in assets_to_copy:
        src_path = preset_package_path / asset.path
        dst_path = root / asset.path
        shutil.copyfile(str(src_path), str(dst_path))
        console.console.print(
            f'Updated [item]{asset.path}[/item] from preset [item]{preset_name}[/item].'
        )


def _try_installing_from_resources(name: str) -> bool:
    if name == LOCAL:
        return False
    rsrc_preset_path = get_default_app_path() / 'presets' / name
    if not rsrc_preset_path.exists():
        return False
    yaml_path = rsrc_preset_path / 'preset.rbx.yml'
    if not yaml_path.is_file():
        return False
    console.console.print(f'Installing preset [item]{name}[/item] from resources...')
    _install(rsrc_preset_path, force=True)
    return True


@functools.cache
def _install_from_resources_just_once(name: str) -> bool:
    # Send all output to the void since we don't wanna be verbose here.
    with console.console.capture():
        return _try_installing_from_resources(name)


def get_installed_preset_or_null(
    name: str, root: pathlib.Path = pathlib.Path()
) -> Optional[Preset]:
    installation_path = get_preset_installation_path(name, root) / 'preset.rbx.yml'
    if not installation_path.is_file():
        if not _try_installing_from_resources(name):
            return None
    elif name == 'default':
        _install_from_resources_just_once(name)

    return utils.model_from_yaml(Preset, installation_path.read_text())


def get_installed_preset(name: str, root: pathlib.Path = pathlib.Path()) -> Preset:
    preset = get_installed_preset_or_null(name, root)
    if preset is None:
        console.console.print(
            f'[error]Preset [item]{name}[/item] is not installed.[/error]'
        )
        raise typer.Exit(1)
    return preset


def optionally_install_environment_from_preset(
    preset: Preset, root: pathlib.Path = pathlib.Path()
):
    if preset.env is None:
        return
    env_path = get_environment_path(preset.name)
    preset_env_path = root / preset.env
    if env_path.is_file():
        if digest_file(preset_env_path) == digest_file(env_path):
            return
        import questionary

        overwrite = questionary.confirm(
            'Preset environment file has changed. Overwrite?',
            default=False,
        ).ask()
        if not overwrite:
            return

    console.console.print(
        f'[success]Overwriting the existing environment based on [item]{preset.env}[/item].'
    )
    env_path.parent.mkdir(parents=True, exist_ok=True)
    shutil.copyfile(str(preset_env_path), env_path)


def _install(root: pathlib.Path = pathlib.Path(), force: bool = False):
    preset = get_preset_yaml(root)

    if preset.name == LOCAL:
        console.console.print('[error]Naming a preset "local" is prohibited.[/error]')

    console.console.print(f'Installing preset [item]{preset.name}[/item]...')
    installation_path = get_preset_installation_path(preset.name)

    if root.resolve().is_relative_to(installation_path.resolve()):
        console.console.print(
            '[error]Current folder is nested into the preset installation path, cannot install it.[/error]'
        )
        raise typer.Exit(1)

    if preset.env is not None:
        console.console.print(
            f'[item]{preset.name}[/item]: Copying environment file...'
        )
        should_copy_env = True
        if get_environment_path(preset.name).exists():
            res = force or rich.prompt.Confirm.ask(
                f'Environment [item]{preset.name}[/item] already exists. Overwrite?',
                console=console.console,
            )
            if not res:
                should_copy_env = False

        if should_copy_env:
            get_environment_path(preset.name).parent.mkdir(parents=True, exist_ok=True)
            shutil.rmtree(get_environment_path(preset.name), ignore_errors=True)
            shutil.copyfile(str(root / preset.env), get_environment_path(preset.name))

    console.console.print(f'[item]{preset.name}[/item]: Copying preset folder...')
    installation_path.parent.mkdir(parents=True, exist_ok=True)
    if installation_path.exists():
        res = force or rich.prompt.Confirm.ask(
            f'Preset [item]{preset.name}[/item] is already installed. Overwrite?',
            console=console.console,
        )
        if not res:
            raise typer.Exit(1)
    shutil.rmtree(str(installation_path), ignore_errors=True)
    shutil.copytree(str(root), str(installation_path))
    shutil.rmtree(str(installation_path / 'build'), ignore_errors=True)
    shutil.rmtree(str(installation_path / '.box'), ignore_errors=True)
    shutil.rmtree(str(installation_path / '.local.rbx'), ignore_errors=True)
    shutil.rmtree(str(installation_path / '.git'), ignore_errors=True)


def install_from_remote(fetch_info: PresetFetchInfo, force: bool = False) -> str:
    import git

    assert fetch_info.fetch_uri is not None
    with tempfile.TemporaryDirectory() as d:
        console.console.print(
            f'Cloning preset from [item]{fetch_info.fetch_uri}[/item]...'
        )
        git.Repo.clone_from(fetch_info.fetch_uri, d)
        pd = pathlib.Path(d)
        if fetch_info.inner_dir:
            pd = pd / fetch_info.inner_dir
            console.console.print(
                f'Installing preset from [item]{fetch_info.inner_dir}[/item].'
            )
        preset = get_preset_yaml(pd)
        preset.uri = fetch_info.uri

        (pd / 'preset.rbx.yml').write_text(utils.model_to_yaml(preset))
        _install(pd, force=force)
        return preset.name


def generate_lock(
    preset_name: Optional[str] = None, root: pathlib.Path = pathlib.Path()
):
    if preset_name is None:
        preset_lock = get_preset_lock(root)
        if preset_lock is None:
            console.console.print(
                '[error][item].preset-lock.yml[/item] not found. '
                'Specify a preset argument to this function to create a lock from scratch.[/error]'
            )
            raise typer.Exit(1)
        preset_name = preset_lock.preset_name

    preset = get_installed_preset(preset_name, root)

    tracked_assets = _get_preset_tracked_assets(preset_name, is_contest=_is_contest())
    preset_lock = PresetLock(
        name=preset.name if preset_name != LOCAL else LOCAL,
        uri=preset.uri,
        assets=_build_package_locked_assets(tracked_assets, root),
    )

    (root / '.preset-lock.yml').write_text(utils.model_to_yaml(preset_lock))
    console.console.print(
        '[success][item].preset-lock.yml[/item] was created.[/success]'
    )


def _sync(try_update: bool = False):
    preset_lock = get_preset_lock()
    if preset_lock is None:
        console.console.print(
            '[error]Package does not have a [item].preset.lock.yml[/item] file and thus cannot be synced.[/error]'
        )
        console.console.print(
            '[error]Ensure this package was created through a preset, or manually associate one with [item]rbx presets lock [PRESET][/item][/error]'
        )
        raise typer.Exit(1)

    should_update = try_update and preset_lock.uri is not None
    installed_preset = get_installed_preset_or_null(preset_lock.preset_name)
    if installed_preset is None:
        if not try_update or preset_lock.uri is None:
            console.console.print(
                f'[error]Preset [item]{preset_lock.preset_name}[/item] is not installed. Install it before trying to update.'
            )
            raise typer.Exit(1)
        install(preset_lock.uri)
    elif should_update:
        update(preset_lock.name)

    _copy_updated_assets(
        preset_lock.preset_name,
        preset_lock,
        is_contest=_is_contest(),
    )
    generate_lock(preset_lock.preset_name)


@app.command(
    'install', help='Install preset from current directory or from the given URI.'
)
def install(
    uri: Optional[str] = typer.Argument(
        None, help='GitHub URI for the preset to install.'
    ),
):
    if uri is None:
        _install()
        return

    fetch_info = get_preset_fetch_info(uri)
    if fetch_info is None:
        console.console.print(f'[error] Preset with URI {uri} not found.[/error]')
        raise typer.Exit(1)
    if fetch_info.fetch_uri is None:
        console.console.print(f'[error]URI {uri} is invalid.[/error]')
    install_from_remote(fetch_info)


@app.command('update', help='Update installed remote presets')
def update(
    name: Annotated[
        Optional[str], typer.Argument(help='If set, update only this preset.')
    ] = None,
):
    if not name:
        presets = _find_installed_presets()
    else:
        presets = [name]

    for preset_name in presets:
        if preset_name == LOCAL:
            import questionary

            if not questionary.confirm(
                'Updating local preset will remove all custom changes you made to the preset.',
                default=False,
            ).ask():
                continue

        preset = get_installed_preset_or_null(preset_name)
        if preset is None:
            console.console.print(
                f'[error]Preset [item]{preset_name}[/item] is not installed.'
            )
            continue
        if preset.uri is None:
            console.console.print(
                f'Skipping preset [item]{preset_name}[/item], not remote.'
            )
            continue
        install_from_remote(preset.fetch_info, force=True)

        if preset_name == LOCAL:
            # Get global path to the preset.
            preset_path = get_preset_installation_path(preset.name)
            dest_path = '.local.rbx'
            shutil.rmtree(dest_path, ignore_errors=True)
            shutil.copytree(preset_path, dest_path)
            console.console.print(
                '[success]Local preset updated successfully.[/success]'
            )
        else:
            console.console.print(
                f'[success]Preset [item]{preset_name}[/item] updated successfully.[/success]'
            )


@app.command(
    'sync',
    help='Sync current package assets with those provided by the installed preset.',
)
@cd.within_closest_package
def sync(
    update: Annotated[
        bool,
        typer.Option(
            '--update',
            '-u',
            help='Whether to fetch an up-to-date version of the installed preset from remote, if available.',
        ),
    ] = False,
):
    _check_is_valid_package()
    _sync(try_update=update)


@app.command(
    'lock', help='Generate a lock for this package, based on a existing preset.'
)
@cd.within_closest_package
def lock(
    preset: Annotated[
        Optional[str],
        typer.Argument(
            help='Preset to generate a lock for. If unset, will default to the one in the existing .preset-lock.yml.',
        ),
    ] = None,
):
    _check_is_valid_package()
    generate_lock(preset)


@app.callback()
def callback():
    pass



import functools
import time
from shutil import rmtree
from typing import List, Optional

import syncer
import typer
from pydantic import BaseModel

from rbx import console
from rbx.box import checkers, package, validators
from rbx.box.code import SanitizationLevel, compile_item, run_item
from rbx.box.generators import (
    GenerationMetadata,
    expand_generator_call,
    generate_standalone,
)
from rbx.box.retries import Retrier
from rbx.box.schema import CodeItem, GeneratorCall, Stress, TaskType, Testcase
from rbx.box.solutions import compile_solutions, get_outcome_style_verdict
from rbx.box.stressing import finder_parser
from rbx.grading.steps import (
    DigestOrDest,
    DigestOrSource,
    Evaluation,
    Outcome,
    TestcaseIO,
    TestcaseLog,
)
from rbx.utils import StatusProgress


class StressFinding(BaseModel):
    generator: GeneratorCall


class StressReport(BaseModel):
    findings: List[StressFinding] = []
    executed: int = 0


def _compile_finder(finder: CodeItem) -> str:
    try:
        digest = compile_item(finder)
    except Exception as e:
        console.console.print(
            f'[error]Failed compiling checker [item]{finder.path}[/item][/error]'
        )
        raise typer.Exit(1) from e
    return digest


async def run_stress(
    name: str,
    timeoutInSeconds: int,
    finder: Optional[str] = None,
    args: Optional[str] = None,
    findingsLimit: int = 1,
    verbose: bool = False,
    progress: Optional[StatusProgress] = None,
    sanitized: bool = False,
) -> StressReport:
    pkg = package.find_problem_package_or_die()
    if pkg.type == TaskType.COMMUNICATION:
        console.console.print(
            '[error]Communication problems do not support stress testing.[/error]'
        )
        raise typer.Exit(1)

    if finder:
        stress = Stress(
            name=f'{name}',
            generator=GeneratorCall(name=name, args=args or ''),
            finder=finder,
        )
    else:
        stress = package.get_stress(name)

    call = stress.generator
    generator = package.get_generator(call.name)

    try:
        generator_digest = compile_item(generator, sanitized=SanitizationLevel.PREFER)
    except:
        console.console.print(
            f'[error]Failed compiling generator [item]{generator.name}[/item].[/error]'
        )
        raise

    # Finder expression parser
    parsed_finder = finder_parser.parse(stress.finder)

    solutions = finder_parser.get_all_solution_items(parsed_finder)
    finders = finder_parser.get_all_checker_items(parsed_finder)
    needs_expected_output = finder_parser.needs_expected_output(parsed_finder)

    solution_indices = {str(solution.path): i for i, solution in enumerate(solutions)}
    solutions_digest = compile_solutions(
        tracked_solutions=set(str(solution.path) for solution in solutions),
        sanitized=sanitized,
    )
    if progress:
        progress.update('Compiling finders...')
    finders_digest = {str(finder.path): _compile_finder(finder) for finder in finders}

    compiled_validator = validators.compile_main_validator()

    # Erase old stress directory
    runs_dir = package.get_problem_runs_dir()
    stress_dir = runs_dir / '.stress'
    rmtree(str(stress_dir), ignore_errors=True)
    stress_dir.mkdir(parents=True, exist_ok=True)
    empty_path = runs_dir / '.stress' / '.empty'
    empty_path.write_text('')

    startTime = time.monotonic()

    executed = 0
    findings = []

    while len(findings) < findingsLimit:
        if time.monotonic() - startTime > timeoutInSeconds:
            break

        executed += 1

        if progress:
            seconds = timeoutInSeconds - int(time.monotonic() - startTime)
            progress.update(
                f'Stress testing: found [item]{len(findings)}[/item] tests, '
                f'executed [item]{executed}[/item], '
                f'[item]{seconds}[/item] second(s) remaining...'
            )

        input_path = runs_dir / '.stress' / 'input'
        input_path.parent.mkdir(parents=True, exist_ok=True)

        expanded_generator_call = expand_generator_call(stress.generator)
        await generate_standalone(
            GenerationMetadata(
                generator_call=expanded_generator_call,
                copied_to=Testcase(inputPath=input_path),
            ),
            generator_digest=generator_digest,
            validator_digest=compiled_validator[1]
            if compiled_validator is not None
            else None,
        )

        @functools.cache
        async def run_solution_fn(
            solution: str,
            retry_index: Optional[int] = None,
            input_path=input_path,
        ) -> TestcaseLog:
            index = solution_indices[solution]
            sol = solutions[index]
            output_path = input_path.with_stem(f'{index}').with_suffix('.out')
            stderr_path = output_path.with_suffix('.err')

            run_log = await run_item(
                sol,
                DigestOrSource.create(solutions_digest[sol.path]),
                stdin=DigestOrSource.create(input_path),
                stdout=DigestOrDest.create(output_path),
                stderr=DigestOrDest.create(stderr_path),
                retry_index=retry_index,
            )

            return TestcaseLog(
                **(run_log.model_dump() if run_log is not None else {}),
                stdout_absolute_path=output_path.absolute(),
                stderr_absolute_path=stderr_path.absolute(),
            )

        # Get main solution output.
        expected_output_path = empty_path
        if needs_expected_output:
            main_testcase_log = await run_solution_fn(str(solutions[0].path))
            main_checker_result = checkers.check_with_no_output(main_testcase_log)
            if main_checker_result.outcome != Outcome.ACCEPTED:
                console.console.print(
                    '[error]Error while generating main solution output.[/error]'
                )
                console.console.print(f'Input written at [item]{input_path}[/item]')
                console.console.print(
                    f'Output written at [item]{main_testcase_log.stdout_absolute_path}[/item]'
                )
                console.console.print(
                    f'Stderr written at [item]{main_testcase_log.stderr_absolute_path}[/item]'
                )
                console.console.print()
                console.console.print(
                    "[warning]If you don't want reference outputs to be generated for the tests, you should "
                    "use the two-way modifier in your finder expression (':2')."
                )
                raise typer.Exit(1)
            expected_output_path = main_testcase_log.stdout_absolute_path

        @functools.cache
        async def run_solution_and_checker_fn(
            call: finder_parser.FinderCall,
            input_path=input_path,
            expected_output_path=expected_output_path,
        ) -> finder_parser.FinderResult:
            async def run_fn(retry_index: int) -> Evaluation:
                solution = call.solution
                checker = call.checker

                testcase_log = await run_solution_fn(solution, retry_index=retry_index)
                assert testcase_log.stdout_absolute_path is not None

                if checker is None:
                    checker_result = checkers.check_with_no_output(testcase_log)
                else:
                    checker_digest = finders_digest[checker.path]
                    checker_result = await checkers.check(
                        checker_digest,
                        testcase_log,
                        Testcase(inputPath=input_path, outputPath=expected_output_path),
                        program_output=testcase_log.stdout_absolute_path,
                    )

                return Evaluation(
                    result=checker_result,
                    testcase=TestcaseIO(
                        index=0,
                        input=input_path,
                        output=expected_output_path,
                    ),
                    log=testcase_log,
                )

            retrier = Retrier(is_stress=True)
            eval = await retrier.repeat(run_fn)

            return finder_parser.FinderResult(
                solution=call.solution,
                outcome=eval.result.outcome,
                checker=call.checker,
                solution_log=eval.log,
                checker_result=eval.result,
            )

        runner = finder_parser.FinderTreeRunner(
            runner=syncer.sync(run_solution_and_checker_fn)
        )
        finder_outcome: finder_parser.FinderOutcome = runner.transform(parsed_finder)

        internal_error_results = [
            result
            for result in finder_outcome.results
            if result.outcome == Outcome.INTERNAL_ERROR
        ]

        if internal_error_results:
            console.console.print(
                f'[error]Checkers failed during stress test [item]{name}[/item] with args [info]{expanded_generator_call.name} {expanded_generator_call.args}[/info][/error]'
            )
            for internal_error_result in internal_error_results:
                assert internal_error_result.checker is not None
                assert internal_error_result.checker_result is not None
                internal_error_checker_name = internal_error_result.checker.path
                console.console.print(
                    f'[warning]Checker [item]{internal_error_checker_name}[/item] failed with message:'
                )
                console.console.print(internal_error_result.checker_result.message)
            raise typer.Exit(1)

        if not finder_outcome.truth_value:
            continue

        findings_dir = stress_dir / 'findings'
        findings_dir.mkdir(parents=True, exist_ok=True)
        finding_index = len(findings)

        finding_path = findings_dir / f'{finding_index}.in'
        finding_path.write_bytes(input_path.read_bytes())

        if progress:
            console.console.print(
                f'[error]FINDING[/error] Generator args are "[status]{expanded_generator_call.name} {expanded_generator_call.args}[/status]"'
            )
            seen_finder_results = set()
            for finder_result in finder_outcome.results:
                style = get_outcome_style_verdict(finder_result.outcome)
                finder_result_key = (finder_result.solution, finder_result.checker)
                if finder_result_key in seen_finder_results:
                    continue
                seen_finder_results.add(finder_result_key)
                finder_result_report_line = f'{finder_result.solution} = [{style}]{finder_result.outcome.name}[/{style}]'
                if finder_result.checker is not None:
                    finder_result_report_line += (
                        f' [item]ON[/item] {finder_result.checker.path}'
                    )
                console.console.print(finder_result_report_line)

        findings.append(
            StressFinding(
                generator=expanded_generator_call,
            )
        )

        # Be cooperative.
        time.sleep(0.001)

    return StressReport(findings=findings, executed=executed)


def print_stress_report(report: StressReport):
    console.console.rule('Stress test report', style='status')
    console.console.print(f'Executed [item]{report.executed}[/item] tests.')
    if not report.findings:
        console.console.print('No stress test findings.')
        return
    console.console.print(f'Found [item]{len(report.findings)}[/item] testcases.')

    findings_dir = package.get_problem_runs_dir() / '.stress' / 'findings'
    console.console.print(f'Findings: {findings_dir.resolve()}')
    console.console.print()

    for i, finding in enumerate(report.findings):
        console.console.print(f'[error]Finding {i + 1}[/error]')
        console.console.print(
            f'Generator: [status]{finding.generator.name} {finding.generator.args}[/status]'
        )
        console.console.print()



import contextlib
import os
import signal
import threading
from typing import Optional, Set

_processing_context_pids: Optional[Set[int]] = None
_lock = threading.Lock()

# Creating a processing context is not thread-safe, but adding to it is.


@contextlib.contextmanager
def new_processing_context():
    global _processing_context_pids
    with _lock:
        old_processing_context_pids = _processing_context_pids
        _processing_context_pids = set()
    try:
        yield
    finally:
        with _lock:
            _processing_context_pids = old_processing_context_pids


def get_processing_context() -> Set[int]:
    with _lock:
        return _processing_context_pids or set()


def add_to_processing_context(pid: int):
    global _processing_context_pids
    with _lock:
        if _processing_context_pids is None:
            return
        _processing_context_pids.add(pid)


def terminate_all_processes_in_context():
    with _lock:
        if _processing_context_pids is None:
            return
        for pid in _processing_context_pids:
            try:
                os.kill(pid, signal.SIGTERM)
            except OSError:
                pass
        _processing_context_pids.clear()



from typing import List, Optional

import syncer

from rbx import console
from rbx.box import checkers, package, validators
from rbx.box.schema import CodeItem, Testcase, ValidatorOutcome, ValidatorTest
from rbx.utils import StatusProgress


def _get_validator_for_test(test: ValidatorTest) -> Optional[CodeItem]:
    pkg = package.find_problem_package_or_die()
    if test.validator is not None:
        return test.validator
    return pkg.validator


async def run_validator_unit_tests(progress: StatusProgress):
    pkg = package.find_problem_package_or_die()

    vals: List[CodeItem] = []
    for test in pkg.unitTests.validator:
        val = _get_validator_for_test(test)
        if val is not None:
            vals.append(val)

    compiled_validators = validators.compile_validators(vals, progress=progress)

    if progress:
        progress.update('Running validator unit tests...')

    console.console.rule('Validator tests', style='info')

    for i, test in enumerate(pkg.unitTests.validator):
        val = _get_validator_for_test(test)
        if val is None:
            console.console.print(
                f'[warning]No validator found for test [item]#{i + 1}[/item], skipping.[/warning]'
            )
            continue

        compiled_digest = compiled_validators[str(val.path)]
        info = await validators.validate_one_off(
            test.input,
            val,
            compiled_digest,
        )

        is_valid = test.outcome == ValidatorOutcome.VALID

        markup = (
            '[success]OK[/success]' if info.ok == is_valid else '[error]FAIL[/error]'
        )

        console.console.print(
            f'{markup} Unit test [item]#{i + 1}[/item] for [item]{test.input}[/item]'
        )
        console.console.print(f'  [status]Expected[/status] {test.outcome.value}')
        if info.ok != is_valid:
            if info.ok:
                console.console.print('  [status]Actual[/status] VALID')
            else:
                console.console.print(f'  [status]Actual[/status] {info.message}')


async def run_checker_unit_tests(progress: StatusProgress):
    pkg = package.find_problem_package_or_die()
    if not pkg.unitTests.checker:
        return

    if not package.get_checker():
        console.console.print(
            '[warning]No checker found, skipping checker unit tests.[/warning]'
        )
        return

    compiled_digest = checkers.compile_checker(progress=progress)

    if progress:
        progress.update('Running checker unit tests...')

    console.console.rule('Checker tests', style='info')

    empty_file = package.get_empty_sentinel_path()

    for i, test in enumerate(pkg.unitTests.checker):
        result = await checkers.check(
            compiled_digest,
            run_log=None,
            testcase=Testcase(
                inputPath=test.input or empty_file,
                outputPath=test.answer or empty_file,
            ),
            program_output=test.output or empty_file,
            skip_run_log=True,
        )

        markup = (
            '[success]OK[/success]'
            if test.outcome.match(result.outcome)
            else '[error]FAIL[/error]'
        )

        console.console.print(f'{markup} Unit test [item]#{i + 1}[/item]')
        console.console.print(f'  [status]Expected[/status] {test.outcome.name}')

        if not test.outcome.match(result.outcome):
            console.console.print(f'  [status]Actual[/status] {result.outcome.name}')
            if result.message:
                console.console.print(f'  [status]Message[/status] {result.message}')


@syncer.sync
async def run_unit_tests(progress: StatusProgress):
    await run_validator_unit_tests(progress)
    await run_checker_unit_tests(progress)



import atexit
import pathlib
import tempfile
from typing import Dict, List, Optional

import syncer
from rich.columns import Columns
from rich.panel import Panel
from rich.progress import MofNCompleteColumn, Progress, SpinnerColumn
from rich.text import Text

from rbx import annotations, config, grading_utils, metadata, testcase_rendering
from rbx.config import Language, get_config
from rbx.console import console, multiline_prompt
from rbx.grading import steps
from rbx.grading.judge.sandbox import SandboxBase
from rbx.grading.judge.sandboxes import stupid_sandbox
from rbx.schema import DumpedProblem, Problem


def get_testcase_index(path: pathlib.Path) -> int:
    return int(path.stem.split('.')[-1])


def get_testcases_io(
    problem: DumpedProblem, root: pathlib.Path = pathlib.Path()
) -> List[steps.TestcaseIO]:
    testcases_per_index: Dict[int, steps.TestcaseIO] = {}
    for input_file in root.glob(f'{problem.code}.*.in'):
        try:
            index = get_testcase_index(input_file)
        except ValueError:
            continue
        testcases_per_index[index] = steps.TestcaseIO(index=index, input=input_file)

    for output_file in root.glob(f'{problem.code}.*.out'):
        index = get_testcase_index(output_file)
        try:
            index = get_testcase_index(output_file)
        except ValueError:
            continue
        if index in testcases_per_index:
            testcases_per_index[index].output = output_file
            continue
        testcases_per_index[index] = steps.TestcaseIO(index=index, output=output_file)

    return sorted(testcases_per_index.values(), key=lambda x: x.index)


async def _run_testcases(
    problem: Problem,
    lang: Language,
    lang_name: Optional[str],
    sandbox: SandboxBase,
    testcases: List[steps.TestcaseIO],
    persist_root: pathlib.Path = pathlib.Path(),
) -> Dict[int, Optional[steps.TestcaseLog]]:
    logs: Dict[int, Optional[steps.TestcaseLog]] = {}

    # Ensure persist dir exists.
    persist_root.mkdir(parents=True, exist_ok=True)

    progress = Progress(
        SpinnerColumn(),
        *Progress.get_default_columns(),
        MofNCompleteColumn(),
        transient=True,
    )
    with progress:
        for testcase in progress.track(testcases, description='Running testcases...'):
            params = grading_utils.build_run_sandbox_params(
                problem, testcase.input is not None
            )
            artifacts = grading_utils.build_run_grading_artifacts(
                testcase, persist_root
            )
            run_log = await steps.run(
                lang.exec,
                params,
                sandbox,
                artifacts,
                metadata=steps.RunLogMetadata(language=lang_name),
            )
            if not run_log:
                logs[testcase.index] = None
                continue
            logs[testcase.index] = steps.TestcaseLog(
                **run_log.__dict__,
                stdout_absolute_path=persist_root / f'stdout-{testcase.index}.txt',
                stderr_absolute_path=persist_root / f'stderr-{testcase.index}.txt',
            )

    return logs


def _evaluate_testcases(
    problem: DumpedProblem,
    sandbox: SandboxBase,
    testcases: List[steps.TestcaseIO],
    testcase_logs: Dict[int, Optional[steps.TestcaseLog]],
    persist_root: pathlib.Path = pathlib.Path(),
) -> List[steps.Evaluation]:
    evaluations = []
    artifacts = grading_utils.build_checker_run_grading_artifacts(
        problem,
        persist_root,
    )
    for testcase in testcases:
        if testcase.index not in testcase_logs:
            continue

        log = testcase_logs[testcase.index]
        evaluations.append(
            steps.evaluate(
                sandbox,
                testcase,
                log,
                artifacts,
                should_use_python_checker=not problem.checker,
            )
        )

    return evaluations


def _pretty_print_output_on_panel(file: Optional[pathlib.Path], title: str) -> Panel:
    if not file:
        return Panel('[error]No file to read from.[/error]', title=title, expand=False)
    return Panel(
        testcase_rendering.render_from_file(file),
        title=title,
        expand=False,
    )


def _pretty_print_side_by_side(result: steps.Evaluation):
    if not result.testcase.output:
        return _pretty_print_output_on_panel(result.log.stdout_absolute_path, 'Output')
    return Columns(
        [
            _pretty_print_output_on_panel(result.testcase.output, 'Expected'),
            _pretty_print_output_on_panel(result.log.stdout_absolute_path, 'Actual'),
        ],
        equal=True,
        expand=False,
    )


def _get_outcome_style(outcome: steps.Outcome) -> str:
    if outcome == steps.Outcome.ACCEPTED:
        return 'success'
    if outcome == steps.Outcome.JUDGE_FAILED or outcome == steps.Outcome.INTERNAL_ERROR:
        return 'warning'
    return 'error'


def _pretty_print_outcome_panel(
    problem: DumpedProblem, eval: steps.Evaluation
) -> Panel:
    result: steps.CheckerResult = eval.result
    is_tle = result.outcome == steps.Outcome.TIME_LIMIT_EXCEEDED or (
        problem.timeLimit and eval.log.time * 1000 > problem.timeLimit
    )

    text = Text()
    text.append('Outcome: ')
    text.append(
        result.outcome.value,
        style=_get_outcome_style(result.outcome),
    )
    text.append(' ' * 4)
    text.append('Time: ')
    text.append(f'{eval.log.time:.2f}s', style='error' if is_tle else 'item')
    text.append('\n')
    if eval.testcase.input:
        text.append(f'Input path: {eval.testcase.input.absolute()}')
        text.append('\n')
    if eval.testcase.output:
        text.append(f'Expected path: {eval.testcase.output.absolute()}')
        text.append('\n')
    text.append(f'Answer path: {eval.log.stdout_absolute_path}')
    return Panel(
        text,
        title=f'[bold]Testcase [item]#{eval.testcase.index}[/item]',
        expand=False,
    )


def _pretty_print_evaluation_result(
    problem: DumpedProblem,
    eval: steps.Evaluation,
    interactive: bool = False,
):
    console.print(_pretty_print_outcome_panel(problem, eval))
    if eval.result.outcome != steps.Outcome.ACCEPTED:
        if interactive:
            console.print(
                _pretty_print_output_on_panel(eval.log.stdout_absolute_path, 'Output')
            )
        else:
            console.print(_pretty_print_side_by_side(eval))
        if eval.result.message:
            console.print(
                f'[error]Checker message:[/error] {eval.result.message.strip()}'
            )
    console.print()


def pretty_print_summary(
    problem: DumpedProblem,
    lang: Language,
    evals: List[steps.Evaluation],
    root: pathlib.Path = pathlib.Path(),
):
    submission_file = root / lang.get_submit_file(problem.code)
    passed = sum(1 for eval in evals if eval.result.outcome == steps.Outcome.ACCEPTED)
    total = len(evals)
    console.print(f'Summary for problem [item]{problem.pretty_name()}[/item]:')

    # Test summary.
    text = Text()
    text.append('Passed tests: ')
    text.append(f'{passed}/{total}', style='success' if passed == total else 'error')
    console.print(text)

    console.print(f'Submission file: {submission_file.absolute()}')


def pretty_print_evaluation_results(
    problem: DumpedProblem,
    evals: List[steps.Evaluation],
    interactive: bool = False,
):
    for eval in evals:
        _pretty_print_evaluation_result(problem, eval, interactive=interactive)


@syncer.sync
async def main(
    problem: annotations.Problem,
    language: annotations.LanguageWithDefault = None,
    keep_sandbox: bool = False,
    interactive: bool = False,
    index: Optional[annotations.TestcaseIndex] = None,
):
    dumped_problem = metadata.find_problem_by_anything(problem)
    if not dumped_problem:
        console.print(
            f'[error]Problem with identifier [item]{problem}[/item] not found.[/error]'
        )
        return

    lang = get_config().get_language(language)
    if not lang:
        console.print(
            f'[error]Language {language or get_config().defaultLanguage} not found in config. Please check your configuration.[/error]'
        )
        return

    if interactive:
        testcases = []
        while True:
            console.print(
                f'Providing IO for testcase [item]#{len(testcases)}[/item]...'
            )
            input = multiline_prompt('Testcase input')
            if not input.strip():
                break
            output = multiline_prompt('Testcase output')
            input_path = pathlib.Path(tempfile.mktemp())
            output_path = pathlib.Path(tempfile.mktemp())
            input_path.write_text(input)
            output_path.write_text(output)
            testcases.append(
                steps.TestcaseIO(
                    index=len(testcases), input=input_path, output=output_path
                )
            )
    else:
        testcases = get_testcases_io(dumped_problem)

    if index is not None:
        testcases = [tc for tc in testcases if tc.index == index]

    if not testcases:
        console.print(
            f'[error]No testcases found for the problem [item]{dumped_problem.pretty_name()}[/item].[/error]'
        )
        return

    box = stupid_sandbox.StupidSandbox()
    atexit.register(lambda: box.cleanup(delete=not keep_sandbox))
    persist_root = config.get_empty_app_persist_path()

    with console.status(
        f'Preprocessing code for problem [item]{dumped_problem.pretty_name()}[/item] in language [item]{language or get_config().defaultLanguage}[/item]...'
    ):
        if lang.preprocess:
            preprocess_cmds = grading_utils.build_preprocess_commands(
                dumped_problem, lang
            )
            sandbox_params = grading_utils.build_preprocess_sandbox_params()
            artifacts = grading_utils.build_compile_grading_artifacts(
                dumped_problem, lang
            )
            if not steps.compile(preprocess_cmds, sandbox_params, box, artifacts):
                console.print(
                    f'[error]Failed to preprocess problem [item]{dumped_problem.pretty_name()}[/item].[/error]'
                )
                return

    with console.status(
        f'Compiling checker for problem [item]{dumped_problem.pretty_name()}[/item]...'
    ):
        command = '/usr/bin/g++ -std=c++17 -o checker checker.cpp'
        artifacts = grading_utils.build_checker_compile_grading_artifacts(
            dumped_problem, persist_root
        )
        if dumped_problem.checker and not steps.compile(
            [command], grading_utils.build_preprocess_sandbox_params(), box, artifacts
        ):
            console.print(
                f'[error]Failed to compile checker for problem [item]{dumped_problem.pretty_name()}[/item].[/error]'
            )
            return

    testcase_logs = await _run_testcases(
        dumped_problem, lang, language, box, testcases, persist_root
    )

    if not testcase_logs:
        console.print(
            f'[error]Failed to run testcases for problem [item]{dumped_problem.pretty_name()}[/item]. Sandbox probably crashed.[/error]'
        )
        return

    with console.status(
        f'Evaluating testcases for problem [item]{dumped_problem.pretty_name()}[/item]...'
    ):
        evals = _evaluate_testcases(
            dumped_problem, box, testcases, testcase_logs, persist_root
        )
    if not evals:
        console.print(
            f'[error]Failed to evaluate testcases for problem [item]{dumped_problem.pretty_name()}[/item].[/error]'
        )
        return
    pretty_print_evaluation_results(dumped_problem, evals, interactive=interactive)
    pretty_print_summary(dumped_problem, lang, evals)



import dataclasses
import os
import pathlib
import resource
import signal
import stat
import sys
from time import monotonic
from typing import Any, Dict, List, Optional, Set, Union


@dataclasses.dataclass
class Options:
    output_file: str
    argv: List[str]
    chdir: Optional[str] = None
    stdin_file: Optional[str] = None
    stdout_file: Optional[str] = None
    stderr_file: Optional[str] = None
    time_limit: Optional[float] = None
    wall_time_limit: Optional[float] = None  # seconds
    memory_limit: Optional[int] = None  # kb, but passed in args as mb
    fs_limit: Optional[int] = None  # kb
    files_to_open: List[int] = dataclasses.field(default_factory=list)
    file_duplicates: Dict[int, List[str]] = dataclasses.field(default_factory=dict)
    prefixed: Set[str] = dataclasses.field(default_factory=set)
    prefix: str = ''


def exit_with(code: int):
    sys.exit(code)


def get_tee_command(files: List[str]) -> str:
    path = (
        os.path.join(os.path.dirname(os.path.realpath(__file__)), 'tee.py')
        + ' '
        + ' '.join(files)
    )
    return sys.executable + ' ' + path


valid_modes = ['a', 'w']


@dataclasses.dataclass
class Tee:
    file: Any
    prefix: Union[str, bytes] = ''


def create_tee(files, mode, buffer_size=4096, prefix=''):
    """Get a file object that will mirror writes across multiple files objs

    Options:
        files     A list of files and/or file objects. All strings will be
                  treated as file paths and opened for writing. Everything
                  else is assumed to be a file-like object that implements
                  both the write() and flush() methods.

        mode      Which mode to use when opening new files. Valid values
                  are 'a' (append) and 'w' (overwrite).

        buffer_size
                Control the size of the buffer between writes to the
                resulting file object and the list of files.
    """
    if mode not in valid_modes:
        raise IOError(
            'Only valid modes to create_tee() are: %s' % ', '.join(valid_modes)
        )

    tee_list = []
    for file in files:
        if isinstance(file, Tee):
            tee_list.append(file)
        else:
            tee_list.append(Tee(file))
    for tee in tee_list:
        if isinstance(tee.file, str):
            tee.file = open(tee.file, f'{mode}b')
        if isinstance(tee.prefix, str):
            tee.prefix = tee.prefix.encode()

    pipe_read, pipe_write = os.pipe()
    pid = os.fork()
    if pid == 0:
        # Child -- Read bytes from the pipe and write them to the specified
        #          files.
        try:
            # Close parent's end of the pipe
            os.close(pipe_write)

            bytes = os.read(pipe_read, buffer_size)
            while bytes:
                for tee in tee_list:
                    if tee.prefix:
                        tee.file.write(tee.prefix)
                    tee.file.write(bytes)
                    tee.file.flush()
                    # TODO maybe add in fsync() here if the fileno() method
                    # exists on file

                bytes = os.read(pipe_read, buffer_size)
        except Exception:
            pass
        finally:
            os._exit(255)
    else:
        # Parent -- Return a file object wrapper around the pipe to the
        #           child.
        return os.fdopen(pipe_write, 'w', closefd=False)


def parse_opts() -> Options:
    options = Options(output_file=sys.argv[1], argv=[])
    options.files_to_open = []
    num_opts = 0
    while num_opts + 2 < len(sys.argv) and sys.argv[num_opts + 2].startswith('-'):
        # Process option
        opt = sys.argv[num_opts + 2]
        if opt.startswith('-t'):
            options.time_limit = float(opt[2:])
        elif opt.startswith('-w'):
            options.wall_time_limit = float(opt[2:])
        elif opt.startswith('-m'):
            options.memory_limit = int(opt[2:]) * 1024
        elif opt.startswith('-i'):
            options.stdin_file = opt[2:]
            options.files_to_open.append(0)
        elif opt.startswith('-o'):
            options.stdout_file = opt[2:]
            options.files_to_open.append(1)
        elif opt.startswith('-e'):
            options.stderr_file = opt[2:]
            options.files_to_open.append(2)
        elif opt.startswith('-d') or opt.startswith('-D'):
            is_prefixed = opt.startswith('-D')
            possibilities = [None, 'o', 'e']
            index = possibilities.index(opt[2])
            if index not in options.file_duplicates:
                options.file_duplicates[index] = []
            options.file_duplicates[index].append(opt[3:])
            if is_prefixed:
                options.prefixed.add(opt[3:])
        elif opt.startswith('-c'):
            options.chdir = opt[2:]
        elif opt.startswith('-f'):
            options.fs_limit = int(opt[2:])
        elif opt.startswith('-P'):
            options.prefix = opt[2:]
        else:
            raise Exception(f'Invalid option {opt}')
        num_opts += 1
    options.argv = sys.argv[num_opts + 2 :]
    return options


def get_memory_usage(ru: resource.struct_rusage) -> int:
    if sys.platform == 'darwin':
        return ru.ru_maxrss // 1024 + ru.ru_ixrss
    return ru.ru_maxrss + ru.ru_ixrss + ru.ru_idrss + ru.ru_isrss


def get_cpu_time(ru: resource.struct_rusage) -> float:
    return ru.ru_utime + ru.ru_stime


def _get_file_size(filename: Optional[str]) -> int:
    if filename is None:
        return 0
    path = pathlib.Path(filename)
    if not path.is_file():
        return 0
    return path.stat().st_size


def get_file_sizes(options: Options):
    return _get_file_size(options.stdout_file) + _get_file_size(options.stderr_file)


def set_rlimits(options: Options):
    if options.time_limit is not None:
        time_limit_in_ms = int(options.time_limit * 1000)
        rlimit_cpu = int((time_limit_in_ms + 999) // 1000)
        resource.setrlimit(resource.RLIMIT_CPU, (rlimit_cpu, rlimit_cpu + 1))
    if options.fs_limit is not None:
        fs_limit = options.fs_limit * 1024  # in bytes
        resource.setrlimit(resource.RLIMIT_FSIZE, (fs_limit + 1, fs_limit * 2))


def redirect_fds(options: Options):
    files = [options.stdin_file, options.stdout_file, options.stderr_file]

    for i in options.files_to_open:
        file = files[i]
        if file is None:
            continue
        open_args = [
            os.O_WRONLY | os.O_TRUNC | os.O_CREAT,
            stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH | stat.S_IWUSR,
        ]
        if i == 0:
            # stdin
            open_args = [os.O_RDONLY]
        if i in options.file_duplicates:
            dups = [
                Tee(f, prefix=options.prefix if f in options.prefixed else '')
                for f in options.file_duplicates[i]
            ]
            tee = create_tee(dups + [file], 'a', prefix=options.prefix)
            fd = tee.fileno()
        else:
            fd = os.open(
                file,
                *open_args,
            )
        os.dup2(fd, i)
        os.close(fd)


def wait_and_finish(
    pid: int,
    options: Options,
    start_time: float,
    status_holder: Set[str],
    alarm_msg: Optional[List[Optional[str]]] = None,
):
    _, exitstatus, ru = os.wait4(pid, 0)
    wall_time = monotonic() - start_time
    cpu_time = get_cpu_time(ru)
    memory_used = get_memory_usage(ru)
    file_sizes = get_file_sizes(options)

    entries = []
    exitcode = os.waitstatus_to_exitcode(exitstatus)
    entries.append(f'exit-code: {exitcode}')
    if exitcode < 0:
        entries.append(f'exit-sig: {-exitcode}')

    status = status_holder
    if exitcode > 0:
        status.add('RE')
    if exitcode < 0:
        status.add('SG')
    if options.time_limit is not None and (
        cpu_time > options.time_limit or -exitcode == 24
    ):
        status.add('TO')
        cpu_time = max(cpu_time, options.time_limit)
    if options.wall_time_limit is not None and wall_time > options.wall_time_limit:
        status.add('WT')
        status.add('TO')
    if options.memory_limit is not None and memory_used > options.memory_limit:
        status.add('ML')
    if options.fs_limit is not None and file_sizes > options.fs_limit * 1024:
        status.add('OL')

    if status:
        status_str = ','.join(status)
        entries.append(f'status: {status_str}')

    if alarm_msg:
        alarm_str = ','.join(msg for msg in alarm_msg if msg is not None)
        if alarm_str:
            entries.append(f'alarm-msg: {alarm_str}')

    entries.append(f'time: {cpu_time:.3f}')
    entries.append(f'time-wall: {wall_time:.3f}')
    entries.append(f'mem: {memory_used}')
    entries.append(f'file: {file_sizes}')

    output_file = pathlib.Path(sys.argv[1])
    output_file.parent.mkdir(parents=True, exist_ok=True)
    output_file.write_text('\n'.join(entries) + '\n')


def main():
    options = parse_opts()

    start_time = monotonic()
    sub_pid = os.fork()
    if sub_pid == 0:
        if options.chdir is not None:
            os.chdir(options.chdir)
        set_rlimits(options)
        redirect_fds(options)
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
        os.execvp(options.argv[0], options.argv)

    alarm_msg: List[Optional[str]] = [None]
    status_holder: Set[str] = set()

    def handle_alarm(*args, **kwargs):
        nonlocal alarm_msg
        wall_time = monotonic() - start_time
        if options.wall_time_limit is not None and wall_time > options.wall_time_limit:
            alarm_msg[0] = 'wall timelimit'
            os.kill(sub_pid, 9)
            return
        ru = resource.getrusage(resource.RUSAGE_CHILDREN)
        if options.time_limit is not None:
            cpu_time = get_cpu_time(ru)
            if cpu_time > options.time_limit:
                alarm_msg[0] = 'timelimit'
                os.kill(sub_pid, 9)
                return
        if options.memory_limit is not None:
            memory_used = get_memory_usage(ru)
            if memory_used > options.memory_limit:
                alarm_msg[0] = 'memorylimit'
                os.kill(sub_pid, 9)
                return

        signal.setitimer(signal.ITIMER_REAL, 0.3)

    def handle_sub_term(*args, **kwargs):
        nonlocal status_holder
        status_holder.add('TE')
        os.kill(sub_pid, 9)

    signal.setitimer(signal.ITIMER_REAL, 0.3)
    signal.signal(signal.SIGALRM, handle_alarm)
    signal.signal(signal.SIGTERM, handle_sub_term)

    wait_and_finish(sub_pid, options, start_time, status_holder, alarm_msg=alarm_msg)
    # Cancel alarm before exiting to avoid surprises.
    signal.setitimer(signal.ITIMER_REAL, 0)

    # Exit gracefully.
    sys.exit(0)


if __name__ == '__main__':
    main()
# type: ignore



import atexit
import os
import shlex

from rbx import annotations, grading_utils, metadata
from rbx.config import get_config
from rbx.console import stderr_console
from rbx.grading import steps
from rbx.grading.judge.sandboxes import stupid_sandbox


def main(
    problem: annotations.Problem,
    language: annotations.LanguageWithDefault = None,
    keep_sandbox: bool = False,
):
    dumped_problem = metadata.find_problem_by_anything(problem)
    if not dumped_problem:
        stderr_console.print(
            f'[error]Problem with identifier [item]{problem}[/item] not found.[/error]'
        )
        return

    lang = get_config().get_language(language)
    if not lang:
        stderr_console.print(
            f'[error]Language {language or get_config().defaultLanguage} not found in config. Please check your configuration.[/error]'
        )
        return

    box = stupid_sandbox.StupidSandbox()
    atexit.register(lambda: box.cleanup(delete=not keep_sandbox))

    preprocess_cmds = grading_utils.build_preprocess_commands(dumped_problem, lang)
    sandbox_params = grading_utils.build_preprocess_sandbox_params()
    artifacts = grading_utils.build_compile_grading_artifacts(dumped_problem, lang)

    if not steps.compile(preprocess_cmds, sandbox_params, box, artifacts):
        stderr_console.print(
            f'[error]Failed to preprocess problem [item]{dumped_problem.pretty_name()}[/item].[/error]'
        )
        return

    cmd = shlex.split(lang.exec)
    os.execv(cmd[0], cmd)



import pathlib
import shlex
import shutil
import sys
import tempfile
from typing import Annotated, Optional

import rich
import rich.prompt
import syncer
import typer

from rbx import annotations, config, console, utils
from rbx.box import (
    cd,
    compile,
    creation,
    download,
    environment,
    generators,
    package,
    presets,
    setter_config,
    state,
    validators,
)
from rbx.box.contest import main as contest
from rbx.box.contest.contest_package import find_contest_yaml
from rbx.box.environment import VerificationLevel, get_environment_path
from rbx.box.packaging import main as packaging
from rbx.box.schema import CodeItem, ExpectedOutcome, TestcaseGroup
from rbx.box.solutions import (
    estimate_time_limit,
    get_exact_matching_solutions,
    get_matching_solutions,
    pick_solutions,
    print_run_report,
    run_and_print_interactive_solutions,
    run_solutions,
)
from rbx.box.statements import build_statements
from rbx.box.testcase_utils import TestcaseEntry
from rbx.box.testcases import main as testcases

app = typer.Typer(no_args_is_help=True, cls=annotations.AliasGroup)
app.add_typer(
    setter_config.app,
    name='config, cfg',
    cls=annotations.AliasGroup,
    help='Manage setter configuration (sub-command).',
    rich_help_panel='Configuration',
)
app.add_typer(
    build_statements.app,
    name='statements, st',
    cls=annotations.AliasGroup,
    help='Manage statements (sub-command).',
    rich_help_panel='Deploying',
)
app.add_typer(
    download.app,
    name='download',
    cls=annotations.AliasGroup,
    help='Download an asset from supported repositories (sub-command).',
    rich_help_panel='Management',
)
app.add_typer(
    presets.app,
    name='presets',
    cls=annotations.AliasGroup,
    help='Manage presets (sub-command).',
    rich_help_panel='Configuration',
)
app.add_typer(
    packaging.app,
    name='package, pkg',
    cls=annotations.AliasGroup,
    help='Build problem packages (sub-command).',
    rich_help_panel='Deploying',
)
app.add_typer(
    contest.app,
    name='contest',
    cls=annotations.AliasGroup,
    help='Manage contests (sub-command).',
    rich_help_panel='Management',
)
app.add_typer(
    testcases.app,
    name='testcases, tc, t',
    cls=annotations.AliasGroup,
    help='Manage testcases (sub-command).',
    rich_help_panel='Management',
)


@app.callback()
def main(
    sanitized: bool = typer.Option(
        False,
        '--sanitized',
        '-s',
        help='Whether to compile and run testlib components with sanitizers enabled. '
        'If you want to run the solutions with sanitizers enabled, use the "-s" flag in the corresponding run command.',
    ),
    debug_logs: bool = typer.Option(
        False,
        '--debug-logs',
        '--debug',
        '-d',
        help='Whether to save extra debug logs along with the evaluation results.',
    ),
):
    state.STATE.run_through_cli = True
    state.STATE.sanitized = sanitized
    if sanitized:
        console.console.print(
            '[warning]Sanitizers are running just for testlib components.\n'
            'If you want to run the solutions with sanitizers enabled, use the [item]-s[/item] flag in the corresponding run command.[/warning]'
        )
    state.STATE.debug_logs = debug_logs


# @app.command('ui', hidden=True)
# @package.within_problem
# def ui():
#     ui_pkg.start()


@app.command(
    'edit, e',
    rich_help_panel='Configuration',
    help='Open problem.rbx.yml in your default editor.',
)
@package.within_problem
def edit():
    console.console.print('Opening problem definition in editor...')
    # Call this function just to raise exception in case we're no in
    # a problem package.
    package.find_problem()
    config.open_editor(package.find_problem_yaml() or pathlib.Path())


@app.command(
    'build, b', rich_help_panel='Deploying', help='Build all tests for the problem.'
)
@package.within_problem
@syncer.sync
async def build(verification: environment.VerificationParam):
    from rbx.box import builder

    await builder.build(verification=verification)


@app.command(
    'run, r',
    rich_help_panel='Testing',
    help='Build and run solution(s).',
)
@package.within_problem
@syncer.sync
async def run(
    verification: environment.VerificationParam,
    solution: Annotated[
        Optional[str],
        typer.Argument(
            help='Path to solution to run. If not specified, will run all solutions.'
        ),
    ] = None,
    outcome: Optional[str] = typer.Option(
        None,
        '--outcome',
        '-o',
        help='Include only solutions whose expected outcomes intersect with this.',
    ),
    check: bool = typer.Option(
        True,
        '--nocheck',
        flag_value=False,
        help='Whether to not build outputs for tests and run checker.',
    ),
    detailed: bool = typer.Option(
        False,
        '--detailed',
        '-d',
        help='Whether to print a detailed view of the tests using tables.',
    ),
    timeit: bool = typer.Option(
        False,
        '--time',
        '-t',
        help='Whether to use estimate a time limit based on accepted solutions.',
    ),
    sanitized: bool = typer.Option(
        False,
        '--sanitized',
        '-s',
        help='Whether to compile the solutions with sanitizers enabled.',
    ),
    choice: bool = typer.Option(
        False,
        '--choice',
        '--choose',
        '-c',
        help='Whether to pick solutions interactively.',
    ),
):
    main_solution = package.get_main_solution()
    if check and main_solution is None:
        console.console.print(
            '[warning]No main solution found, running without checkers.[/warning]'
        )
        check = False

    tracked_solutions = None
    if outcome is not None:
        tracked_solutions = {
            str(solution.path)
            for solution in get_matching_solutions(ExpectedOutcome(outcome))
        }
    if solution:
        tracked_solutions = {solution}

    if choice:
        tracked_solutions = set(await pick_solutions(tracked_solutions))
        if not tracked_solutions:
            console.console.print('[error]No solutions selected. Exiting.[/error]')
            raise typer.Exit(1)

    from rbx.box import builder

    if not await builder.build(verification=verification, output=check):
        return

    if verification <= VerificationLevel.VALIDATE.value:
        console.console.print(
            '[warning]Verification level is set to [item]validate (-v1)[/item], so rbx only build tests and validated them.[/warning]'
        )
        return

    override_tl = None
    if timeit:
        if sanitized:
            console.console.print(
                '[error]Sanitizers are known to be time-hungry, so they cannot be used for time estimation.\n'
                'Remove either the [item]-s[/item] flag or the [item]-t[/item] flag to run solutions without sanitizers.[/error]'
            )
            raise typer.Exit(1)

        # Never use sanitizers for time estimation.
        override_tl = await _time_impl(check=check, detailed=False)
        if override_tl is None:
            raise typer.Exit(1)

    if sanitized:
        console.console.print(
            '[warning]Sanitizers are running, so the time limit for the problem will be dropped, '
            'and the environment default time limit will be used instead.[/warning]'
        )

    if sanitized and tracked_solutions is None:
        console.console.print(
            '[warning]Sanitizers are running, and no solutions were specified to run. Will only run [item]ACCEPTED[/item] solutions.'
        )
        tracked_solutions = {
            str(solution.path)
            for solution in get_exact_matching_solutions(ExpectedOutcome.ACCEPTED)
        }

    with utils.StatusProgress('Running solutions...') as s:
        solution_result = run_solutions(
            progress=s,
            tracked_solutions=tracked_solutions,
            check=check,
            verification=VerificationLevel(verification),
            timelimit_override=override_tl,
            sanitized=sanitized,
        )

    console.console.print()
    console.console.rule('[status]Run report[/status]', style='status')
    await print_run_report(
        solution_result,
        console.console,
        VerificationLevel(verification),
        detailed=detailed,
        skip_printing_limits=sanitized,
    )


async def _time_impl(check: bool, detailed: bool) -> Optional[int]:
    if package.get_main_solution() is None:
        console.console.print(
            '[warning]No main solution found, so cannot estimate a time limit.[/warning]'
        )
        return None

    verification = VerificationLevel.ALL_SOLUTIONS.value

    with utils.StatusProgress('Running ACCEPTED solutions...') as s:
        tracked_solutions = {
            str(solution.path)
            for solution in get_exact_matching_solutions(ExpectedOutcome.ACCEPTED)
        }
        solution_result = run_solutions(
            progress=s,
            tracked_solutions=tracked_solutions,
            check=check,
            verification=VerificationLevel(verification),
            timelimit_override=600,  # 10 minute time limit for estimation
        )

    console.console.print()
    console.console.rule(
        '[status]Run report (for time estimation)[/status]', style='status'
    )
    ok = await print_run_report(
        solution_result,
        console.console,
        VerificationLevel(verification),
        detailed=detailed,
        skip_printing_limits=True,
    )

    if not ok:
        console.console.print(
            '[error]Failed to run ACCEPTED solutions, so cannot estimate a reliable time limit.[/error]'
        )
        return None

    console.console.print()
    return await estimate_time_limit(console.console, solution_result)


@app.command(
    'time, t',
    rich_help_panel='Testing',
    help='Estimate a time limit for the problem based on a time limit formula and timings of accepted solutions.',
)
@package.within_problem
@syncer.sync
async def time(
    check: bool = typer.Option(
        True,
        '--nocheck',
        flag_value=False,
        help='Whether to not build outputs for tests and run checker.',
    ),
    detailed: bool = typer.Option(
        False,
        '--detailed',
        '-d',
        help='Whether to print a detailed view of the tests using tables.',
    ),
):
    main_solution = package.get_main_solution()
    if check and main_solution is None:
        console.console.print(
            '[warning]No main solution found, running without checkers.[/warning]'
        )
        check = False

    from rbx.box import builder

    verification = VerificationLevel.ALL_SOLUTIONS.value
    if not await builder.build(verification=verification, output=check):
        return None

    await _time_impl(check, detailed)


@app.command(
    'irun, ir',
    rich_help_panel='Testing',
    help='Build and run solution(s) by passing testcases in the CLI.',
)
@package.within_problem
@syncer.sync
async def irun(
    verification: environment.VerificationParam,
    solution: Annotated[
        Optional[str],
        typer.Argument(
            help='Path to solution to run. If not specified, will run all solutions.'
        ),
    ] = None,
    outcome: Optional[str] = typer.Option(
        None,
        '--outcome',
        '-o',
        help='Include only solutions whose expected outcomes intersect with this.',
    ),
    check: bool = typer.Option(
        True,
        '--nocheck',
        flag_value=False,
        help='Whether to not build outputs for tests and run checker.',
    ),
    generator: Optional[str] = typer.Option(
        None,
        '--generator',
        '-g',
        help='Generator call to use to generate a single test for execution.',
    ),
    testcase: Optional[str] = typer.Option(
        None,
        '--testcase',
        '--test',
        '-tc',
        '-t',
        help='Testcase to run, in the format "[group]/[index]". If not specified, will run interactively.',
    ),
    output: bool = typer.Option(
        False,
        '--output',
        '-o',
        help='Whether to ask user for custom output.',
    ),
    print: bool = typer.Option(
        False, '--print', '-p', help='Whether to print outputs to terminal.'
    ),
    sanitized: bool = typer.Option(
        False,
        '--sanitized',
        '-s',
        help='Whether to compile the solutions with sanitizers enabled.',
    ),
    choice: bool = typer.Option(
        False,
        '--choice',
        '--choose',
        '-c',
        help='Whether to pick solutions interactively.',
    ),
):
    if not print:
        console.console.print(
            '[warning]Outputs will be written to files. If you wish to print them to the terminal, use the "-p" parameter.'
        )
    if verification < VerificationLevel.ALL_SOLUTIONS.value:
        console.console.print(
            '[warning]Verification level should be at least [item]all solutions (-v4)[/item] to run solutions interactively.'
        )
        return

    tracked_solutions = None
    if outcome is not None:
        tracked_solutions = {
            str(solution.path)
            for solution in get_matching_solutions(ExpectedOutcome(outcome))
        }
    if solution:
        tracked_solutions = {solution}

    if choice:
        tracked_solutions = set(await pick_solutions(tracked_solutions))
        if not tracked_solutions:
            console.console.print('[error]No solutions selected. Exiting.[/error]')
            raise typer.Exit(1)

    if sanitized and tracked_solutions is None:
        console.console.print(
            '[warning]Sanitizers are running, and no solutions were specified to run. Will only run [item]ACCEPTED[/item] solutions.'
        )
        tracked_solutions = {
            str(solution.path)
            for solution in get_exact_matching_solutions(ExpectedOutcome.ACCEPTED)
        }

    with utils.StatusProgress('Running solutions...') as s:
        await run_and_print_interactive_solutions(
            progress=s,
            tracked_solutions=tracked_solutions,
            check=check,
            verification=VerificationLevel(verification),
            generator=generators.get_call_from_string(generator)
            if generator is not None
            else None,
            testcase_entry=TestcaseEntry.parse(testcase) if testcase else None,
            custom_output=output,
            print=print,
            sanitized=sanitized,
        )


@app.command(
    'create, c',
    rich_help_panel='Management',
    help='Create a new problem package.',
)
def create(
    name: str,
    preset: Annotated[
        Optional[str], typer.Option(help='Preset to use when creating the problem.')
    ] = None,
):
    if find_contest_yaml() is not None:
        console.console.print(
            '[error]Cannot [item]rbx create[/item] a problem inside a contest.[/error]'
        )
        console.console.print(
            '[error]Instead, use [item]rbx contest add[/item] to add a problem to a contest.[/error]'
        )
        raise typer.Exit(1)

    if preset is not None:
        creation.create(name, preset=preset)
        return
    creation.create(name)


@app.command(
    'stress',
    rich_help_panel='Testing',
    help='Run a stress test.',
)
@package.within_problem
@syncer.sync
async def stress(
    name: Annotated[
        str,
        typer.Argument(
            help='Name of the stress test to run (specified in problem.rbx.yml), '
            'or the generator to run, in case -g is specified.'
        ),
    ],
    generator_args: Annotated[
        Optional[str],
        typer.Option(
            '--generator',
            '-g',
            help='Run generator [name] with these args.',
        ),
    ] = None,
    finder: Annotated[
        Optional[str],
        typer.Option(
            '--finder',
            '-f',
            help='Run a stress with this finder expression.',
        ),
    ] = None,
    timeout: Annotated[
        int,
        typer.Option(
            '--timeout',
            '--time',
            '-t',
            help='For how many seconds to run the stress test.',
        ),
    ] = 10,
    findings: Annotated[
        int,
        typer.Option('--findings', '-n', help='How many breaking tests to look for.'),
    ] = 1,
    verbose: bool = typer.Option(
        False,
        '-v',
        '--verbose',
        help='Whether to print verbose output for checkers and finders.',
    ),
    sanitized: bool = typer.Option(
        False,
        '--sanitized',
        '-s',
        help='Whether to compile the solutions with sanitizers enabled.',
    ),
):
    if finder and not generator_args or generator_args and not finder:
        console.console.print(
            '[error]Options --generator/-g and --finder/-f should be specified together.'
        )
        raise typer.Exit(1)

    from rbx.box import stresses

    with utils.StatusProgress('Running stress...') as s:
        report = await stresses.run_stress(
            name,
            timeout,
            args=generator_args,
            finder=finder,
            findingsLimit=findings,
            progress=s,
            verbose=verbose,
            sanitized=sanitized,
        )

    stresses.print_stress_report(report)

    if not report.findings:
        return

    # Add found tests.
    res = rich.prompt.Confirm.ask(
        'Do you want to add the tests that were found to a test group?',
        console=console.console,
    )
    if not res:
        return
    testgroup = None
    while testgroup is None or testgroup:
        groups_by_name = {
            name: group
            for name, group in package.get_test_groups_by_name().items()
            if group.generatorScript is not None
            and group.generatorScript.path.suffix == '.txt'
        }

        import questionary

        testgroup = await questionary.select(
            'Choose the testgroup to add the tests to.\nOnly test groups that have a .txt generatorScript are shown below: ',
            choices=list(groups_by_name) + ['(create new script)', '(skip)'],
        ).ask_async()

        if testgroup == '(create new script)':
            new_script_name = await questionary.text(
                'Enter the name of the new .txt generatorScript file: '
            ).ask_async()
            new_script_path = pathlib.Path(new_script_name).with_suffix('.txt')
            new_script_path.parent.mkdir(parents=True, exist_ok=True)
            new_script_path.touch()

            # Temporarily create a new testgroup with the new script.
            testgroup = new_script_path.stem
            groups_by_name[testgroup] = TestcaseGroup(
                name=testgroup, generatorScript=CodeItem(path=new_script_path)
            )
            ru, problem_yml = package.get_ruyaml()
            if 'testcases' not in problem_yml:
                problem_yml['testcases'] = []
            problem_yml['testcases'].append(
                {
                    'name': testgroup,
                    'generatorScript': new_script_path.name,
                }
            )
            dest = package.find_problem_yaml()
            assert dest is not None
            utils.save_ruyaml(dest, ru, problem_yml)
            package.clear_package_cache()

        if testgroup not in groups_by_name:
            break
        try:
            subgroup = groups_by_name[testgroup]
            assert subgroup.generatorScript is not None
            generator_script = pathlib.Path(subgroup.generatorScript.path)

            finding_lines = []
            for finding in report.findings:
                line = finding.generator.name
                if finding.generator.args is not None:
                    line = f'{line} {finding.generator.args}'
                finding_lines.append(line)

            with generator_script.open('a') as f:
                stress_text = f'# Obtained by running `rbx {shlex.join(sys.argv[1:])}`'
                finding_text = '\n'.join(finding_lines)
                f.write(f'\n{stress_text}\n{finding_text}\n')

            console.console.print(
                f"Added [item]{len(report.findings)}[/item] tests to test group [item]{testgroup}[/item]'s generatorScript at [item]{subgroup.generatorScript.path}[/item]"
            )
        except typer.Exit:
            continue
        break


@app.command(
    'compile',
    rich_help_panel='Testing',
    help='Compile an asset given its path.',
)
@package.within_problem
@syncer.sync
async def compile_command(
    path: Annotated[
        Optional[str],
        typer.Argument(help='Path to the asset to compile.'),
    ] = None,
    sanitized: bool = typer.Option(
        False,
        '--sanitized',
        '-s',
        help='Whether to compile the asset with sanitizers enabled.',
    ),
    warnings: bool = typer.Option(
        False,
        '--warnings',
        '-w',
        help='Whether to compile the asset with warnings enabled.',
    ),
):
    if path is None:
        import questionary

        path = await questionary.path("What's the path to your asset?").ask_async()
        if path is None:
            console.console.print('[error]No path specified.[/error]')
            raise typer.Exit(1)

    compile.any(path, sanitized, warnings)


@app.command(
    'validate',
    rich_help_panel='Testing',
    help='Run the validator in a one-off fashion, interactively.',
)
@package.within_problem
@syncer.sync
async def validate(
    path: Annotated[
        Optional[str],
        typer.Option('--path', '-p', help='Path to the testcase to validate.'),
    ] = None,
):
    validator_tuple = validators.compile_main_validator()
    if validator_tuple is None:
        console.console.print('[error]No validator found for this problem.[/error]')
        raise typer.Exit(1)

    validator, validator_digest = validator_tuple

    input = console.multiline_prompt('Testcase input')

    if path is None:
        with tempfile.TemporaryDirectory() as tmpdir:
            tmppath = pathlib.Path(tmpdir) / '000.in'
            tmppath.write_text(input)

            info = await validators.validate_one_off(
                pathlib.Path(tmppath), validator, validator_digest
            )
    else:
        info = await validators.validate_one_off(
            pathlib.Path(path), validator, validator_digest
        )

    validators.print_validation_report([info])


@app.command(
    'unit',
    rich_help_panel='Testing',
    help='Run unit tests for the validator and checker.',
)
def unit_tests():
    from rbx.box import unit

    with utils.StatusProgress('Running unit tests...') as s:
        unit.run_unit_tests(s)


@app.command(
    'environment, env',
    rich_help_panel='Configuration',
    help='Set or show the current box environment.',
)
def environment_command(
    env: Annotated[Optional[str], typer.Argument()] = None,
    install_from: Annotated[
        Optional[str],
        typer.Option(
            '--install',
            '-i',
            help='Whether to install this environment from the given file.',
        ),
    ] = None,
):
    if env is None:
        cfg = config.get_config()
        console.console.print(f'Current environment: [item]{cfg.boxEnvironment}[/item]')
        console.console.print(
            f'Location: {environment.get_environment_path(cfg.boxEnvironment)}'
        )
        return
    if install_from is not None:
        environment.install_environment(env, pathlib.Path(install_from))
    if not get_environment_path(env).is_file():
        console.console.print(
            f'[error]Environment [item]{env}[/item] does not exist.[/error]'
        )
        raise typer.Exit(1)

    cfg = config.get_config()
    if env == cfg.boxEnvironment:
        console.console.print(
            f'Environment is already set to [item]{env}[/item].',
        )
        return
    console.console.print(
        f'Changing box environment from [item]{cfg.boxEnvironment}[/item] to [item]{env}[/item]...'
    )
    cfg.boxEnvironment = env
    config.save_config(cfg)

    # Also clear cache when changing environments.
    clear()


@app.command(
    'activate',
    rich_help_panel='Configuration',
    help='Activate the environment of the current preset used by the package.',
)
@cd.within_closest_package
def activate():
    preset_lock = presets.get_preset_lock()
    if preset_lock is None:
        console.console.print(
            '[warning]No configured preset to be activated for this package.[/warning]'
        )
        raise typer.Exit(1)

    preset = presets.get_installed_preset_or_null(preset_lock.preset_name)
    if preset is None:
        if preset_lock.uri is None:
            console.console.print(
                '[error]Preset is not installed. Install it manually, or specify a URI in [item].preset-lock.yml[/item].[/error]'
            )
            raise typer.Exit(1)
        presets.install(preset_lock.uri)

    preset = presets.get_installed_preset(preset_lock.preset_name)

    # Install the environment from the preset if it's not already installed.
    presets.optionally_install_environment_from_preset(
        preset, root=presets.get_preset_installation_path(preset_lock.name)
    )

    # Activate the environment.
    if preset.env is not None:
        environment_command(preset.name)

    console.console.print(f'[success]Preset [item]{preset.name}[/item] is activated.')


@app.command(
    'languages',
    rich_help_panel='Configuration',
    help='List the languages available in this environment',
)
def languages():
    env = environment.get_environment()

    console.console.print(
        f'[success]There are [item]{len(env.languages)}[/item] language(s) available.'
    )

    for language in env.languages:
        console.console.print(
            f'[item]{language.name}[/item], aka [item]{language.readable_name or language.name}[/item]:'
        )
        console.console.print(language)
        console.console.print()


@app.command(
    'clear, clean',
    rich_help_panel='Management',
    help='Clears cache and build directories.',
)
@cd.within_closest_package
def clear():
    console.console.print('Cleaning cache and build directories...')
    shutil.rmtree('.box', ignore_errors=True)
    shutil.rmtree('build', ignore_errors=True)



import pathlib
import shutil
from typing import List, Optional, Tuple

import typer
from pydantic import BaseModel

from rbx import console
from rbx.box import package
from rbx.box.package import get_build_testgroup_path, get_build_tests_path
from rbx.box.schema import Testcase, TestcaseGroup


class TestcaseEntry(BaseModel):
    group: str
    index: int

    def key(self) -> Tuple[str, int]:
        return self.group, self.index

    def __str__(self) -> str:
        return f'{self.group}/{self.index}'

    @classmethod
    def parse(cls, spec: str) -> 'TestcaseEntry':
        if spec.count('/') != 1:
            console.console.print(
                f'[error]Invalid testcase spec [item]{spec}[/item]. Format should be [item]<group>/<index>[/item].[/error]',
            )
            raise typer.Exit(1)
        group, index = spec.split('/')
        return TestcaseEntry(group=group.strip(), index=int(index))


class TestcasePattern(BaseModel):
    group_prefix: List[str]
    index: Optional[int] = None

    def group(self) -> str:
        return '/'.join(self.group_prefix)

    def match(self, group_entry: TestcaseEntry) -> bool:
        # TODO: support subgroups.
        entry_parts = tuple(group_entry.group.split('/'))
        if self.index is not None:
            if self.index != group_entry.index:
                return False
            if tuple(self.group_prefix) != entry_parts:
                return False
            return True

        if len(self.group_prefix) > len(entry_parts):
            return False

        return tuple(self.group_prefix) == entry_parts[: len(self.group_prefix)]

    def with_no_index(self) -> 'TestcasePattern':
        return self.model_copy(update={'index': None})

    def intersecting_group(self, group: str) -> bool:
        if self.with_no_index().match(TestcaseEntry(group=group, index=0)):
            # If the group is inside the pattern, then it is a match.
            return True
        if TestcasePattern.parse(group).match(
            TestcaseEntry(group=self.group(), index=0)
        ):
            # If the group is a prefix of the pattern, then it is a match.
            return True
        return False

    def __str__(self) -> str:
        prefix = '/'.join(self.group_prefix)
        if not prefix:
            return '*'
        if self.index is None:
            return f'{prefix}/'
        return f'{prefix}/{self.index}'

    @classmethod
    def parse(cls, spec: str) -> 'TestcasePattern':
        spec = spec.strip()
        if spec == '*':
            return cls(group_prefix=[], index=None)

        parts = spec.split('/')
        if len(parts) <= 1:
            return cls(group_prefix=parts, index=None)

        if parts[-1].isdigit():
            return cls(group_prefix=parts[:-1], index=int(parts[-1]))

        return cls(group_prefix=parts, index=None)


class TestcaseData(BaseModel):
    input: str
    output: str


class TestcaseInteractionEntry(BaseModel):
    data: str
    pipe: int


class TestcaseInteraction(BaseModel):
    entries: List[TestcaseInteractionEntry]
    prefixes: Tuple[str, str]


def find_built_testcases(group: TestcaseGroup) -> List[Testcase]:
    inputs = find_built_testcase_inputs(group)

    testcases = []
    for input in inputs:
        output = input.with_suffix('.out')
        testcases.append(Testcase(inputPath=input, outputPath=output))
    return testcases


def find_built_testcase_inputs(group: TestcaseGroup) -> List[pathlib.Path]:
    testgroup_path = get_build_testgroup_path(group.name)
    if not testgroup_path.is_dir():
        console.console.print(
            f'Testgroup {group.name} is not generated in build folder'
        )
        raise typer.Exit(1)

    return sorted(testgroup_path.glob('*.in'))


def clear_built_testcases():
    shutil.rmtree(str(get_build_tests_path()), ignore_errors=True)


def get_samples() -> List[Testcase]:
    tcs = find_built_testcases(package.get_testgroup('samples'))
    return [
        Testcase(
            inputPath=tc.inputPath.resolve(),
            outputPath=tc.outputPath.resolve()
            if tc.outputPath is not None and tc.outputPath.is_file()
            else None,
        )
        for tc in tcs
    ]


def fill_output_for_defined_testcase(testcase: Testcase) -> Testcase:
    res = testcase.model_copy()
    if res.outputPath is not None:
        return res
    output_path = res.inputPath.with_suffix('.ans')
    if output_path.is_file():
        res.outputPath = output_path
    return res


def parse_interaction(file: pathlib.Path) -> TestcaseInteraction:
    entries = []
    with file.open('r') as f:
        try:
            interactor_prefix = f.readline().strip()
            solution_prefix = f.readline().strip()
        except Exception:
            console.console.print(
                f'[error]Failed to read interaction file [item]{file}[/item]. Expected the first two lines to be the interactor and solution prefixes.[/error]'
            )
            raise typer.Exit(1) from None

        rest = f.read()
        start = 0

        def _find_next_prefix(start: int) -> Optional[Tuple[int, int]]:
            interactor_idx = rest.find(interactor_prefix, start)
            solution_idx = rest.find(solution_prefix, start)
            if interactor_idx == -1 and solution_idx == -1:
                return None
            if interactor_idx == -1:
                return (solution_idx, solution_idx + len(solution_prefix))
            if solution_idx == -1:
                return (interactor_idx, interactor_idx + len(interactor_prefix))
            if interactor_idx < solution_idx:
                return (interactor_idx, interactor_idx + len(interactor_prefix))
            return (solution_idx, solution_idx + len(solution_prefix))

        def _find_next_block() -> Optional[Tuple[int, Tuple[int, int]]]:
            prefix = _find_next_prefix(start)
            if prefix is None:
                return None
            prefix_start, prefix_end = prefix
            prefix = rest[prefix_start:prefix_end]
            pipe = 1 if prefix == solution_prefix else 0

            nxt = _find_next_prefix(prefix_end)
            if nxt is None:
                return (pipe, (prefix_end, len(rest)))
            nxt_start, _ = nxt
            return (pipe, (prefix_end, nxt_start))

        while True:
            block = _find_next_block()
            if block is None:
                break
            pipe, (st, nd) = block
            entries.append(TestcaseInteractionEntry(data=rest[st:nd], pipe=pipe))
            start = nd

    return TestcaseInteraction(
        prefixes=(interactor_prefix, solution_prefix),
        entries=entries,
    )



import pathlib
import shutil
from typing import List

import typer

from rbx import console
from rbx.box import package
from rbx.box.environment import get_extension_or_default
from rbx.box.packaging.boca.extension import BocaExtension, BocaLanguage
from rbx.box.packaging.boca.packager import BocaPackager
from rbx.box.packaging.packager import BuiltStatement
from rbx.box.schema import ExpectedOutcome, TaskType
from rbx.config import get_default_app_path, get_testlib
from rbx.grading.judge.digester import digest_cooperatively


class MojPackager(BocaPackager):
    def __init__(self, for_boca: bool = False):
        super().__init__()
        self.for_boca = for_boca

    @classmethod
    def task_types(cls) -> List[TaskType]:
        return [TaskType.COMMUNICATION, TaskType.BATCH]

    def _get_problem_info(self) -> str:
        statement = self._get_main_statement()
        return (
            f'basename={self._get_problem_name()}\n'
            f'fullname={statement.title}\n'
            f'descfile={self._get_problem_name()}.pdf\n'
        )

    def _get_tl(self) -> str:
        extension = get_extension_or_default('boca', BocaExtension)

        pkg = package.find_problem_package_or_die()
        res = f'TL[default]={pkg.timeLimit / 1000}\n'
        for language in extension.languages:
            res += f'TL[{language}]={self._get_pkg_timelimit(language) / 1000}\n'
        return res

    def _get_limits(self) -> str:
        pkg = package.find_problem_package_or_die()
        ml = pkg.memoryLimit
        # ol = pkg.outputLimit
        limits = [
            f'ULIMITS[-v]={ml * 1024}',
            # f'ULIMITS[-f]={ol}',
        ]
        return '\n'.join(limits) + '\n'

    def _get_compare(self) -> str:
        extension = get_extension_or_default('boca', BocaExtension)

        compare_path = (
            get_default_app_path() / 'packagers' / 'moj' / 'scripts' / 'compare.sh'
        )
        if not compare_path.exists():
            console.console.print(
                '[error]MOJ template compare script not found.[/error]'
            )
            raise typer.Exit(1)
        with package.get_checker().path.open('rb') as f:
            checker_hash = digest_cooperatively(f)
        return (
            compare_path.read_text()
            .replace('{{rbxFlags}}', extension.flags_with_defaults()['cc'])
            .replace('{{checkerHash}}', checker_hash)
        )

    def _get_checker(self) -> str:
        return package.get_checker().path.read_text()

    def _get_interactor(self) -> str:
        return package.get_interactor().path.read_text()

    def _expand_language_vars(self, language: BocaLanguage, dir: pathlib.Path):
        extension = get_extension_or_default('boca', BocaExtension)

        for path in dir.glob('**/*'):
            if not path.is_file():
                continue

            replaced = path.read_text()
            replaced = replaced.replace(
                '{{rbxMaxMemory}}', f'{self._get_pkg_memorylimit(language)}'
            ).replace(
                '{{rbxInitialMemory}}',
                f'{min(512, int(self._get_pkg_memorylimit(language) * 0.9))}',
            )

            flags = extension.flags_with_defaults()
            if language in flags:
                replaced = replaced.replace('{{rbxFlags}}', flags[language])

            path.write_text(replaced)

            if path.suffix == '.sh':
                path.chmod(0o755)

    def _copy_solutions_moj(self, into_path: pathlib.Path):
        into_path = into_path / 'sols'
        has_good = False
        for solution in package.get_solutions():
            tag = 'wrong'
            if solution.outcome == ExpectedOutcome.ACCEPTED:
                tag = 'good'
                has_good = True
            elif solution.outcome.is_slow():
                tag = 'slow'
            dest_path = into_path / tag / solution.path.name
            dest_path.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(str(solution.path), dest_path)

        if not has_good:
            console.console.print('[error]No good solution found.[/error]')
            raise typer.Exit(1)

    @classmethod
    def name(cls) -> str:
        return 'moj'

    def package(
        self,
        build_path: pathlib.Path,
        into_path: pathlib.Path,
        built_statements: List[BuiltStatement],
    ) -> pathlib.Path:
        pkg = package.find_problem_package_or_die()

        # Prepare dummy files
        author_path = into_path / 'author'
        author_path.parent.mkdir(parents=True, exist_ok=True)
        author_path.write_text('Unknown\n')

        tags_path = into_path / 'tags'
        tags_path.parent.mkdir(parents=True, exist_ok=True)
        tags_path.write_text('')

        # Prepare limits
        limits_path = into_path / 'conf'
        limits_path.parent.mkdir(parents=True, exist_ok=True)
        limits_path.write_text(self._get_limits())

        # Prepare TL
        if self.for_boca:
            tl_path = into_path / 'tl'
            tl_path.parent.mkdir(parents=True, exist_ok=True)
            tl_path.write_text(self._get_tl())

        # Prepare compare
        compare_path = into_path / 'scripts' / 'compare.sh'
        compare_path.parent.mkdir(parents=True, exist_ok=True)
        compare_path.write_text(self._get_compare())
        compare_path.chmod(0o755)

        # Prepare testlib
        testlib_path = into_path / 'scripts' / 'testlib.h'
        testlib_path.parent.mkdir(parents=True, exist_ok=True)
        testlib_path.write_text(get_testlib().read_text())

        # Prepare checker
        checker_path = into_path / 'scripts' / 'checker.cpp'
        checker_path.parent.mkdir(parents=True, exist_ok=True)
        checker_path.write_text(self._get_checker())

        # Prepare interactor
        if pkg.type == TaskType.COMMUNICATION:
            interactor_path = into_path / 'scripts' / 'interactor.cpp'
            interactor_path.parent.mkdir(parents=True, exist_ok=True)
            interactor_path.write_text(self._get_interactor())

            interactor_prep_path = into_path / 'scripts' / 'interactor_prep.sh'
            interactor_prep_path.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(
                get_default_app_path()
                / 'packagers'
                / 'moj'
                / 'scripts'
                / 'interactor_prep.sh',
                interactor_prep_path,
            )
            interactor_prep_path.chmod(0o755)

            interactor_run_path = into_path / 'scripts' / 'interactor_run.sh'
            interactor_run_path.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(
                get_default_app_path()
                / 'packagers'
                / 'moj'
                / 'scripts'
                / 'interactor_run.sh',
                interactor_run_path,
            )
            interactor_run_path.chmod(0o755)

        # Prepare language scripts
        extension = get_extension_or_default('boca', BocaExtension)
        for language in extension.languages:
            language_path = into_path / 'scripts' / language
            language_path.parent.mkdir(parents=True, exist_ok=True)
            src_path = (
                get_default_app_path() / 'packagers' / 'moj' / 'scripts' / language
            )
            if src_path.exists():
                shutil.copytree(src_path, language_path)
                self._expand_language_vars(language, language_path)

        # Problem statement
        enunciado_path = into_path / 'docs' / 'enunciado.pdf'
        enunciado_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copyfile(
            self._get_main_built_statement(built_statements).path,
            enunciado_path,
        )

        # Copy solutions
        if self.for_boca:
            self._copy_solutions(into_path, fix_java=False)
        else:
            self._copy_solutions_moj(into_path)

        # Prepare IO
        inputs_path = into_path / 'tests' / 'input'
        inputs_path.mkdir(parents=True, exist_ok=True)
        outputs_path = into_path / 'tests' / 'output'
        outputs_path.mkdir(parents=True, exist_ok=True)

        testcases = self.get_flattened_built_testcases()
        for i, testcase in enumerate(testcases):
            shutil.copyfile(testcase.inputPath, inputs_path / f'{i + 1:03d}')
            if testcase.outputPath is not None:
                shutil.copyfile(testcase.outputPath, outputs_path / f'{i + 1:03d}')
            else:
                (outputs_path / f'{i + 1:03d}').touch()

        # Zip all.
        shutil.make_archive(
            str(build_path / self._get_problem_name()), 'zip', into_path
        )

        return (build_path / self._get_problem_name()).with_suffix('.zip')
